# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
from typing import IO, Callable, List, MutableMapping, NamedTuple, Optional, overload
import logging
import os
import select
import subprocess
import sys

if sys.version_info >= (3, 8):
    from typing import Literal  # pragma: no cover
else:  # pragma: no cover
    from typing_extensions import Literal  # pragma: no cover

# project
from kiwi.utils.codec import Codec

from kiwi.exceptions import (
    KiwiCommandError,
    KiwiCommandNotFound
)

log = logging.getLogger('kiwi')


class CommandT(NamedTuple):
    output: str
    error: str
    returncode: int


class CommandCallT(NamedTuple):
    output: IO[bytes]
    output_available: Callable[[], bool]
    error: IO[bytes]
    error_available: Callable[[], bool]
    process: subprocess.Popen


class Command:
    """
    **Implements command invocation**

    An instance of Command provides methods to invoke external
    commands in blocking and non blocking mode. Control of
    stdout and stderr is given to the caller
    """

    @overload
    @staticmethod
    def run(
        command: List[str], custom_env: Optional[MutableMapping[str, str]] = None,
        raise_on_error: bool = True, stderr_to_stdout: bool = False,
        raise_on_command_not_found: Literal[False] = False
    ) -> CommandT:
        ...  # pragma: no cover

    @overload
    @staticmethod
    def run(
        command: List[str], custom_env: Optional[MutableMapping[str, str]] = None,
        raise_on_error: bool = True, stderr_to_stdout: bool = False,
        raise_on_command_not_found: bool = True
    ) -> Optional[CommandT]:
        ...  # pragma: no cover

    @staticmethod
    def run(
        command: List[str], custom_env: Optional[MutableMapping[str, str]] = None,
        raise_on_error: bool = True, stderr_to_stdout: bool = False,
        raise_on_command_not_found: bool = True
    ) -> Optional[CommandT]:
        """
        Execute a program and block the caller. The return value
        is a CommandT namedtuple containing the stdout, stderr
        and return code information. Unless raise_on_error is
        set to `False` an exception is thrown if the command
        exits with an error code not equal to zero. If
        raise_on_command_not_found is `False` and the command is
        not found, then `None` is returned.

        Example:

        .. code:: python

            result = Command.run(['ls', '-l'])

        :param list command: command and arguments
        :param dict custom_env: custom os.environ
        :param bool raise_on_error: control error behaviour
        :param bool stderr_to_stdout: redirects stderr to stdout

        :return:
            Contains call results in command type

            .. code:: python

                CommandT(output='string', error='string', returncode=int)

        :rtype: CommandT
        """
        from .path import Path
        log.debug('EXEC: [%s]', ' '.join(command))
        environment = custom_env or os.environ
        cmd_abspath: Optional[str]
        if command[0].startswith("/"):
            cmd_abspath = command[0]
            if not os.path.exists(cmd_abspath):
                cmd_abspath = None
        else:
            cmd_abspath = Path.which(
                command[0], custom_env=environment, access_mode=os.X_OK
            )

        if not cmd_abspath:
            message = 'Command "%s" not found in the environment' % command[0]
            if raise_on_command_not_found:
                raise KiwiCommandNotFound(message)
            log.debug('EXEC: %s', message)
            return None
        stderr = subprocess.STDOUT if stderr_to_stdout else subprocess.PIPE
        try:
            process = subprocess.Popen(
                [cmd_abspath] + command[1:],
                stdout=subprocess.PIPE,
                stderr=stderr,
                env=environment
            )
        except (OSError, subprocess.SubprocessError) as e:
            raise KiwiCommandError(
                '%s: %s: %s' % (command[0], type(e).__name__, format(e))
            ) from e

        output, error = process.communicate()
        if process.returncode != 0 and raise_on_error:
            if not error:
                error = bytes(b'(no output on stderr)')
            if not output:
                output = bytes(b'(no output on stdout)')
            log.debug(
                'EXEC: Failed with stderr: {0}, stdout: {1}'.format(
                    Codec.decode(error), Codec.decode(output)
                )
            )
            raise KiwiCommandError(
                '{0}: stderr: {1}, stdout: {2}'.format(
                    command[0], Codec.decode(error), Codec.decode(output)
                )
            )
        return CommandT(
            output=Codec.decode(output),
            error=Codec.decode(error),
            returncode=process.returncode
        )

    @staticmethod
    def call(
            command: List[str],
            custom_env: Optional[MutableMapping[str, str]] = None
    ) -> CommandCallT:
        """
        Execute a program and return an io file handle pair back.
        stdout and stderr are both on different channels. The caller
        must read from the output file handles in order to actually
        run the command. This can be done using the CommandIterator
        from command_process

        Example:

        .. code:: python

            process = Command.call(['ls', '-l'])

        :param list command: command and arguments
        :param list custom_env: custom os.environ

        :return:
            Contains process results in command type

            .. code:: python

                command(
                    output='string', output_available=bool,
                    error='string', error_available=bool,
                    process=subprocess
                )

        :rtype: namedtuple
        """
        from .path import Path
        log.debug('EXEC: [%s]', ' '.join(command))
        environment = custom_env or os.environ
        if not Path.which(
            command[0], custom_env=environment, access_mode=os.X_OK
        ):
            raise KiwiCommandNotFound(
                'Command "%s" not found in the environment' % command[0]
            )
        try:
            process = subprocess.Popen(
                command,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                env=environment
            )
        except Exception as e:
            raise KiwiCommandError(
                '%s: %s' % (type(e).__name__, format(e))
            ) from e

        # guaranteed to be true as stdout & stderr equal subprocess.PIPE
        assert process.stdout and process.stderr

        def output_available() -> Callable[[], bool]:
            def _select():
                readable, _, exceptional = select.select(
                    [process.stdout], [], [process.stdout], 1e-4
                )
                if readable and not exceptional:
                    return True
                return False
            return _select

        def error_available() -> Callable[[], bool]:
            def _select():
                readable, _, exceptional = select.select(
                    [process.stderr], [], [process.stderr], 1e-4
                )
                if readable and not exceptional:
                    return True
                return False
            return _select

        return CommandCallT(
            output=process.stdout,
            output_available=output_available(),
            error=process.stderr,
            error_available=error_available(),
            process=process
        )



from builtins import bytes
import encodings.ascii as encoding
from unittest.mock import (
    patch, call, Mock, mock_open
)
from pytest import raises

from kiwi.utils.checksum import Checksum

from kiwi.exceptions import KiwiFileNotFound


class TestChecksum:
    @patch('os.path.exists')
    def setup(self, mock_exists):
        self.ascii = encoding.getregentry().name
        read_results = [bytes(b''), bytes(b'data'), bytes(b''), bytes(b'data')]

        def side_effect(arg):
            print(read_results[0])
            return read_results.pop()

        self.m_open = mock_open()
        self.m_open.return_value.read.side_effect = side_effect

        mock_exists.return_value = True
        self.checksum = Checksum('some-file')

    @patch('os.path.exists')
    def setup_method(self, cls, mock_exists):
        self.setup()

    def test_checksum_file_not_found(self):
        with raises(KiwiFileNotFound):
            Checksum('some-file')

    @patch('os.path.exists')
    def test_matches_checksum_file_does_not_exist(self, mock_exists):
        mock_exists.return_value = False
        assert self.checksum.matches('sum', 'some-file') is False

    @patch('os.path.exists')
    def test_matches(self, mock_exists):
        mock_exists.return_value = True

        self.m_open.return_value.read.side_effect = None
        self.m_open.return_value.read.return_value = 'sum'
        with patch('builtins.open', self.m_open, create=True):
            assert self.checksum.matches('sum', 'some-file') is True

        self.m_open.assert_called_once_with(
            'some-file', encoding=self.ascii
        )

        with patch('builtins.open', self.m_open, create=True):
            assert self.checksum.matches('foo', 'some-file') is False

    @patch('kiwi.path.Path.which')
    @patch('kiwi.utils.checksum.Compress')
    @patch('hashlib.md5')
    @patch('os.path.getsize')
    def test_md5_xz(self, mock_size, mock_md5, mock_compress, mock_which):
        checksum = Mock
        checksum.uncompressed_filename = 'some-file-uncompressed'
        mock_which.return_value = 'factor'
        compress = Mock()
        digest = Mock()
        digest.block_size = 1024
        digest._calculate_hash_hexdigest = Mock(
            return_value=checksum
        )
        digest.hexdigest = Mock(
            return_value='sum'
        )
        compress.get_format = Mock(
            return_value='xz'
        )
        mock_size.return_value = 1343225856
        mock_md5.return_value = digest
        mock_compress.return_value = compress

        with patch('builtins.open', self.m_open, create=True):
            self.checksum.md5('outfile')

        assert self.m_open.call_args_list == [
            call('some-file', 'rb'),
            call('some-file-uncompressed', 'rb'),
            call('outfile', encoding=self.ascii, mode='w')
        ]
        self.m_open.return_value.write.assert_called_once_with(
            'sum 163968 8192 163968 8192\n'
        )

    @patch('kiwi.path.Path.which')
    @patch('kiwi.utils.checksum.Compress')
    @patch('hashlib.md5')
    @patch('os.path.getsize')
    def test_md5(
        self, mock_size, mock_md5, mock_compress, mock_which
    ):
        mock_which.return_value = 'factor'
        compress = Mock()
        digest = Mock()
        digest.block_size = 1024
        digest.hexdigest = Mock(
            return_value='sum'
        )
        compress.get_format = Mock(
            return_value=None
        )
        mock_size.return_value = 1343225856
        mock_md5.return_value = digest
        mock_compress.return_value = compress

        with patch('builtins.open', self.m_open, create=True):
            self.checksum.md5('outfile')

        assert self.m_open.call_args_list == [
            call('some-file', 'rb'),
            call('outfile', encoding=self.ascii, mode='w')
        ]
        self.m_open.return_value.write.assert_called_once_with(
            'sum 163968 8192\n'
        )

    @patch('kiwi.path.Path.which')
    @patch('kiwi.utils.checksum.Compress')
    @patch('hashlib.sha256')
    @patch('os.path.getsize')
    def test_sha256(
        self, mock_size, mock_sha256, mock_compress, mock_which
    ):
        mock_which.return_value = 'factor'
        compress = Mock()
        digest = Mock()
        digest.block_size = 1024
        digest.hexdigest = Mock(
            return_value='sum'
        )
        compress.get_format = Mock(
            return_value=None
        )
        mock_size.return_value = 1343225856
        mock_sha256.return_value = digest
        mock_compress.return_value = compress

        with patch('builtins.open', self.m_open, create=True):
            self.checksum.sha256('outfile')

        assert self.m_open.call_args_list == [
            call('some-file', 'rb'),
            call('outfile', encoding=self.ascii, mode='w')
        ]
        self.m_open.return_value.write.assert_called_once_with(
            'sum 163968 8192\n'
        )

    @patch('hashlib.sha256')
    def test_sha256_plain(self, mock_sha256):
        digest = Mock()
        digest.block_size = 1024
        digest.hexdigest = Mock(
            return_value='sum'
        )
        mock_sha256.return_value = digest

        with patch('builtins.open', self.m_open, create=True):
            assert self.checksum.sha256() == digest.hexdigest.return_value

    @patch('hashlib.md5')
    def test_md5_plain(self, mock_md5):
        digest = Mock()
        digest.block_size = 1024
        digest.hexdigest = Mock(
            return_value='sum'
        )
        mock_md5.return_value = digest

        with patch('builtins.open', self.m_open, create=True):
            assert self.checksum.md5() == digest.hexdigest.return_value



import io
from pytest import raises
from unittest.mock import (
    Mock, patch, call, MagicMock
)
from kiwi.bootloader.config.systemd_boot import BootLoaderSystemdBoot

from kiwi.exceptions import (
    KiwiTemplateError,
    KiwiBootLoaderTargetError,
    KiwiKernelLookupError
)


class TestBootLoaderSystemdBoot:
    @patch('kiwi.bootloader.config.bootloader_spec_base.FirmWare')
    def setup(self, mock_FirmWare):
        self.state = Mock()
        self.state.xml_data.get_name.return_value = 'image-name'
        self.state.get_image_version.return_value = 'image-version'
        self.state.build_type.get_efifatimagesize.return_value = None
        self.bootloader = BootLoaderSystemdBoot(self.state, 'root_dir')
        self.bootloader.custom_args['kernel'] = None
        self.bootloader.custom_args['initrd'] = None
        self.bootloader.custom_args['boot_options'] = {}
        self.bootloader.root_mount = Mock(
            mountpoint='system_root_mount'
        )
        self.bootloader._mount_system = Mock()
        self.bootloader.create_efi_path = Mock()
        self.bootloader.get_boot_path = Mock(
            return_value='bootpath'
        )
        self.bootloader.get_menu_entry_title = Mock(
            return_value='title'
        )

    @patch('kiwi.bootloader.config.bootloader_spec_base.FirmWare')
    def setup_method(self, cls, mock_FirmWare):
        self.setup()

    def test_setup_loader_raises_invalid_target(self):
        with raises(KiwiBootLoaderTargetError):
            self.bootloader.setup_loader('iso')

    @patch('os.path.exists')
    @patch('kiwi.bootloader.config.systemd_boot.OsRelease')
    @patch('kiwi.bootloader.config.systemd_boot.glob.iglob')
    @patch('kiwi.bootloader.config.systemd_boot.BootImageBase.get_boot_names')
    @patch('kiwi.bootloader.config.systemd_boot.Path.wipe')
    @patch('kiwi.bootloader.config.systemd_boot.Path.create')
    @patch('kiwi.bootloader.config.systemd_boot.Command.run')
    @patch('kiwi.bootloader.config.systemd_boot.BootLoaderTemplateSystemdBoot')
    @patch.object(BootLoaderSystemdBoot, '_write_config_file')
    @patch.object(BootLoaderSystemdBoot, '_get_template_parameters')
    @patch.object(BootLoaderSystemdBoot, '_write_kernel_cmdline_file')
    def test_setup_loader(
        self, mock_write_kernel_cmdline_file,
        mock_get_template_parameters, mock_write_config_file,
        mock_BootLoaderTemplateSystemdBoot, mock_Command_run,
        mock_Path_create, mock_Path_wipe, mock_BootImageBase_get_boot_names,
        mock_iglob,
        mock_OsRelease,
        mock_os_path_exists
    ):
        mock_iglob.return_value = ['/lib/modules/5.3.18-59.10-default']
        os_release = Mock()
        os_release.get.return_value = 'opensuse-leap'
        mock_OsRelease.return_value = os_release
        kernel_info = Mock()
        kernel_info.kernel_version = 'kernel-version'
        kernel_info.kernel_filename = 'kernel-filename'
        kernel_info.initrd_name = 'initrd-name'
        mock_os_path_exists.return_value = True
        mock_BootImageBase_get_boot_names.return_value = kernel_info
        self.bootloader.setup_loader('disk')
        assert mock_write_config_file.call_args_list == [
            call(
                mock_BootLoaderTemplateSystemdBoot.
                return_value.get_loader_template.return_value,
                'system_root_mount/boot/efi/loader/loader.conf',
                mock_get_template_parameters.return_value
            ),
            call(
                mock_BootLoaderTemplateSystemdBoot.
                return_value.get_entry_template.return_value,
                'system_root_mount/boot/efi/loader/entries/'
                'opensuse-leap-5.3.18-59.10-default.conf',
                mock_get_template_parameters.return_value
            )
        ]

        assert self.bootloader._mount_system.called
        mock_write_kernel_cmdline_file.assert_called_once_with(
            'system_root_mount'
        )
        assert mock_Command_run.call_args_list == [
            call(
                [
                    'chroot', 'system_root_mount',
                    'bootctl', 'install',
                    '--esp-path=/boot/efi',
                    '--no-variables',
                    '--entry-token', 'os-id'
                ]
            ),
            call(
                [
                    'cp', 'kernel-filename',
                    'system_root_mount/boot/efi/os'
                ]
            ),
            call(
                [
                    'cp', 'system_root_mount/boot/initrd-name',
                    'system_root_mount/boot/efi/os'
                ]
            )
        ]

    @patch('kiwi.bootloader.config.systemd_boot.OsRelease')
    @patch('kiwi.bootloader.config.systemd_boot.glob.iglob')
    @patch.object(BootLoaderSystemdBoot, '_write_config_file')
    def test_set_loader_entry(
        self, mock_write_config_file, mock_iglob, mock_OsRelease
    ):
        mock_iglob.return_value = ['/lib/modules/5.3.18-59.10-default']
        os_release = Mock()
        os_release.get.return_value = 'opensuse-leap'
        mock_OsRelease.return_value = os_release
        self.bootloader.set_loader_entry('root_dir', 'disk')

    @patch('kiwi.bootloader.config.systemd_boot.OsRelease')
    @patch('kiwi.bootloader.config.systemd_boot.glob.iglob')
    def test_set_loader_entry_kernel_lookup_raises(
        self, mock_iglob, mock_OsRelease
    ):
        mock_iglob.return_value = None
        with raises(KiwiKernelLookupError):
            self.bootloader.set_loader_entry('root_dir', 'disk')

    def test_create_loader_image(self):
        self.bootloader.create_loader_image('disk')
        self.bootloader.create_efi_path.assert_called_once_with()

    def test_get_template_parameters(self):
        self.bootloader.timeout = 0
        assert self.bootloader._get_template_parameters() == {
            'kernel_file': 'vmlinuz',
            'initrd_file': 'initrd',
            'boot_options': '',
            'boot_timeout': 0,
            'bootpath': 'bootpath',
            'title': 'title',
            'default_entry': 'main.conf'
        }

    @patch('kiwi.bootloader.config.systemd_boot.Path.create')
    def test_write_config_file(self, mock_Path_create):
        template = Mock()
        template.substitute.return_value = 'data'
        with patch('builtins.open', create=True) as mock_open:
            mock_open.return_value = MagicMock(spec=io.IOBase)
            file_handle = mock_open.return_value.__enter__.return_value
            self.bootloader._write_config_file(template, 'path/some-file', {})
            mock_Path_create.assert_called_once_with('path')
            mock_open.assert_called_once_with('path/some-file', 'w')
            file_handle.write.assert_called_once_with('data')

    def test_write_config_file_raises(self):
        template = Mock()
        template.substitute.side_effect = Exception
        with raises(KiwiTemplateError):
            self.bootloader._write_config_file(template, 'some-file', {})

    def test_write_kernel_cmdline_file(self):
        with patch('builtins.open', create=True) as mock_open:
            mock_open.return_value = MagicMock(spec=io.IOBase)
            file_handle = mock_open.return_value.__enter__.return_value
            self.bootloader._write_kernel_cmdline_file('system_root_mount')
            mock_open.assert_called_once_with(
                'system_root_mount/etc/kernel/cmdline', 'w'
            )
            file_handle.write.assert_called_once_with(self.bootloader.cmdline)

    @patch('kiwi.bootloader.config.systemd_boot.Path.create')
    @patch('kiwi.bootloader.config.systemd_boot.Command.run')
    @patch('kiwi.bootloader.config.systemd_boot.LoopDevice')
    @patch('kiwi.bootloader.config.systemd_boot.Disk')
    @patch('kiwi.bootloader.config.systemd_boot.MountManager')
    @patch.object(BootLoaderSystemdBoot, '_run_bootctl')
    @patch.object(BootLoaderSystemdBoot, 'set_loader_entry')
    def test_create_embedded_fat_efi_image(
        self, mock_set_loader_entry, mock_run_bootctl, mock_MountManager,
        mock_Disk, mock_LoopDevice, mock_Command_run, mock_Path_create
    ):
        target = Mock()
        self.bootloader.target = target
        mock_Disk.return_value.__enter__.return_value.partition_map = {
            'efi': 'efi_device'
        }
        self.bootloader._create_embedded_fat_efi_image('ESP')
        assert mock_MountManager.call_args_list == [
            call(device='efi_device', mountpoint='root_dir/boot/efi'),
            call(device='/dev', mountpoint='root_dir/dev'),
            call(device='/proc', mountpoint='root_dir/proc'),
            call(device='/sys', mountpoint='root_dir/sys')
        ]
        assert mock_Command_run.call_args_list == [
            call(['qemu-img', 'create', 'ESP', '20M']),
            call(['sgdisk', '-n', ':1.0', '-t', '1:EF00', 'ESP']),
            call(
                ['mkdosfs', '-n', 'BOOT', 'efi_device']
            ),
            call(['dd', 'if=efi_device', 'of=ESP.img']),
            call(['mv', 'ESP.img', 'ESP'])
        ]
        mock_run_bootctl.assert_called_once_with('root_dir')
        mock_set_loader_entry.assert_called_once_with('root_dir', target.live)



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
# project
import importlib
from typing import Dict
from abc import (
    ABCMeta, abstractmethod
)
from kiwi.defaults import Defaults
from kiwi.xml_state import XMLState

from kiwi.exceptions import KiwiRequestedTypeError


class ImageBuilder(metaclass=ABCMeta):
    """
    Image builder factory
    """
    @abstractmethod
    def __init__(self) -> None:
        return None  # pragma: no cover

    @staticmethod
    def new(
        xml_state: XMLState, target_dir: str,
        root_dir: str, custom_args: Dict = None
    ):
        image_type = xml_state.get_build_type_name()
        if image_type in Defaults.get_filesystem_image_types():
            name_token = ('filesystem', 'FileSystemBuilder')
        elif image_type in Defaults.get_disk_image_types():
            name_token = ('disk', 'DiskBuilder')
        elif image_type in Defaults.get_live_image_types():
            name_token = ('live', 'LiveImageBuilder')
        elif image_type in Defaults.get_kis_image_types():
            name_token = ('kis', 'KisBuilder')
        elif image_type in Defaults.get_archive_image_types():
            name_token = ('archive', 'ArchiveBuilder')
        elif image_type in Defaults.get_container_image_types():
            name_token = ('container', 'ContainerBuilder')
        else:
            name_token = ('None', 'None')
        try:
            (builder_namespace, builder_name) = name_token
            builder = importlib.import_module(
                'kiwi.builder.{0}'.format(builder_namespace)
            )
            return builder.__dict__[builder_name](
                xml_state, target_dir, root_dir, custom_args
            )
        except Exception as issue:
            raise KiwiRequestedTypeError(
                f'Requested image type {image_type} not supported: {issue}'
            )



import os
import errno
import logging
from pytest import fixture
from stat import ST_MODE
from unittest.mock import patch

from kiwi.utils.sync import DataSync


class TestDataSync:
    @fixture(autouse=True)
    def inject_fixtures(self, caplog):
        self._caplog = caplog

    def setup(self):
        self.sync = DataSync('source_dir', 'target_dir')

    def setup_method(self, cls):
        self.setup()

    @patch('kiwi.utils.sync.Command.run')
    @patch('kiwi.utils.sync.DataSync.target_supports_extended_attributes')
    @patch('os.chmod')
    @patch('os.stat')
    def test_sync_data(
        self, mock_stat, mock_chmod, mock_xattr_support, mock_command
    ):
        mock_stat.return_value = os.stat('.')
        mock_xattr_support.return_value = False
        with self._caplog.at_level(logging.WARNING):
            self.sync.sync_data(
                options=[
                    '--archive', '--hard-links', '--xattrs',
                    '--acls', '--one-file-system'
                ],
                exclude=['exclude_me']
            )
            mock_command.assert_called_once_with(
                [
                    'rsync', '--archive', '--hard-links', '--one-file-system',
                    '--exclude', '/exclude_me', 'source_dir', 'target_dir'
                ]
            )
            mock_chmod.assert_called_once_with(
                'target_dir', mock_stat.return_value[ST_MODE]
            )

    @patch('kiwi.utils.sync.Command.run')
    @patch('os.chmod')
    @patch('os.stat')
    def test_sync_data_force_trailing_slash(
        self, mock_stat, mock_chmod, mock_command
    ):
        mock_stat.return_value = os.stat('.')
        self.sync.sync_data(force_trailing_slash=True)
        mock_command.assert_called_once_with(
            ['rsync', 'source_dir/', 'target_dir']
        )

    @patch('os.getxattr')
    def test_target_supports_extended_attributes(self, mock_getxattr):
        assert self.sync.target_supports_extended_attributes() is True
        mock_getxattr.assert_called_once_with(
            'target_dir', 'user.mime_type'
        )

    @patch('os.getxattr')
    def test_target_does_not_support_extended_attributes(self, mock_getxattr):
        effect = OSError()
        effect.errno = errno.ENOTSUP
        mock_getxattr.side_effect = effect
        assert self.sync.target_supports_extended_attributes() is False



from unittest.mock import (
    patch, Mock
)
from pytest import raises

from kiwi.volume_manager import VolumeManager

from kiwi.exceptions import KiwiVolumeManagerSetupError


class TestVolumeManager:
    def test_volume_manager_not_implemented(self):
        with raises(KiwiVolumeManagerSetupError):
            VolumeManager.new('foo', Mock(), 'root_dir', [Mock()])

    @patch('kiwi.volume_manager.lvm.VolumeManagerLVM')
    def test_volume_manager_lvm(self, mock_lvm):
        device_map = Mock()
        volumes = [Mock()]
        VolumeManager.new('lvm', device_map, 'root_dir', volumes)
        mock_lvm.assert_called_once_with(
            device_map, 'root_dir', volumes, None
        )

    @patch('kiwi.volume_manager.btrfs.VolumeManagerBtrfs')
    def test_volume_manager_btrfs(self, mock_btrfs):
        device_map = Mock()
        volumes = [Mock()]
        VolumeManager.new('btrfs', device_map, 'root_dir', volumes)
        mock_btrfs.assert_called_once_with(
            device_map, 'root_dir', volumes, None
        )



import io
from pytest import raises
from unittest.mock import (
    patch, MagicMock
)

from kiwi.container.setup.appx import ContainerSetupAppx
from kiwi.defaults import Defaults

from kiwi.exceptions import KiwiContainerSetupError


class TestContainerSetupAppx:
    @patch('os.path.exists')
    def setup(self, mock_exists):
        mock_exists.return_value = True
        self.appx = ContainerSetupAppx(
            'root_dir', {
                'metadata_path': '../data/appx',
                'image_version': '1.99.1',
                'history': {
                    'created_by': 'SUSE',
                    'author': 'Marcus',
                    'application_id': '123',
                    'package_version': '2003.12.0.0',
                    'comment': 'Some Information',
                    'launcher': 'openSUSE-Tumbleweed.exe'
                }
            }
        )

    @patch('os.path.exists')
    def setup_method(self, cls, mock_exists):
        self.setup()

    def test_setup_raises_no_manifest_file(self):
        with patch('os.path.exists', return_value=True):
            appx = ContainerSetupAppx(
                'root_dir', {'metadata_path': 'artificial'}
            )
            with patch('os.path.exists', return_value=False):
                with raises(KiwiContainerSetupError):
                    appx.setup()

    @patch('lxml.etree.parse')
    def test_setup_raises_xml_operations(self, mock_parse):
        mock_parse.side_effect = Exception
        with raises(KiwiContainerSetupError):
            self.appx.setup()

    def test_setup(self):
        Defaults.set_platform_name('x86_64')
        with patch('builtins.open', create=True) as mock_open:
            mock_open.return_value = MagicMock(spec=io.IOBase)
            file_handle = mock_open.return_value.__enter__.return_value
            self.appx.setup()
            # we expect all @..@ templates to got replaced
            assert '@' not in file_handle.write.call_args_list



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
from typing import (
    List, Dict, no_type_check
)

# project
from kiwi.repository.base import RepositoryBase
from kiwi.api_helper import decommissioned


class RepositoryDnf(RepositoryBase):
    """
    decommissioned
    """
    @decommissioned
    def post_init(self, custom_args: List = []) -> None:
        pass  # pragma: no cover

    @decommissioned
    def setup_package_database_configuration(self) -> None:
        pass  # pragma: no cover

    @decommissioned
    def use_default_location(self) -> None:
        pass  # pragma: no cover

    @no_type_check
    @decommissioned
    def runtime_config(self) -> Dict:
        pass  # pragma: no cover

    @decommissioned
    def add_repo(
        self, name: str, uri: str, repo_type: str = 'rpm-md',
        prio: int = None, dist: str = None, components: str = None,
        user: str = None, secret: str = None, credentials_file: str = None,
        repo_gpgcheck: bool = False, pkg_gpgcheck: bool = False,
        sourcetype: str = None, use_for_bootstrap: bool = False,
        customization_script: str = None
    ) -> None:
        pass  # pragma: no cover

    @decommissioned
    def import_trusted_keys(self, signing_keys: List) -> None:
        pass  # pragma: no cover

    @decommissioned
    def delete_repo(self, name: str) -> None:
        pass  # pragma: no cover

    @decommissioned
    def delete_all_repos(self) -> None:
        pass  # pragma: no cover

    @decommissioned
    def delete_repo_cache(self, name: str) -> None:
        pass  # pragma: no cover

    @decommissioned
    def cleanup_unused_repos(self) -> None:
        pass  # pragma: no cover

    @no_type_check
    @decommissioned
    def _create_dnf_runtime_environment(self) -> Dict:
        pass  # pragma: no cover

    @decommissioned
    def _create_runtime_config_parser(self) -> None:
        pass  # pragma: no cover

    @decommissioned
    def _create_runtime_plugin_config_parsers(self) -> None:
        pass  # pragma: no cover

    @decommissioned
    def _write_runtime_config(self) -> None:
        pass  # pragma: no cover



import logging
from unittest.mock import (
    patch, call, Mock
)
from pytest import (
    raises, fixture
)

from kiwi.package_manager.apt import PackageManagerApt

import kiwi.defaults as defaults

from kiwi.exceptions import (
    KiwiDebootstrapError,
    KiwiRequestError,
    KiwiFileNotFound
)


class TestPackageManagerApt:
    @fixture(autouse=True)
    def inject_fixtures(self, caplog):
        self._caplog = caplog

    def setup(self):
        repository = Mock()
        repository.root_dir = 'root-dir'
        repository.signing_keys = ['key-file.asc']
        repository.keyring = 'trusted.gpg'
        repository.unauthenticated = 'false'
        repository.components = ['main', 'restricted']

        repository.runtime_config = Mock(
            return_value={
                'apt_get_args': ['-c', 'apt.conf', '-y'],
                'command_env': ['env'],
                'distribution': 'xenial',
                'distribution_path': 'xenial_path'
            }
        )
        self.manager = PackageManagerApt(repository)

    def setup_method(self, cls):
        self.setup()

    def test_request_package(self):
        self.manager.request_package('name')
        assert self.manager.package_requests == ['name']

    def test_request_collection(self):
        self.manager.request_collection('name')
        with self._caplog.at_level(logging.WARNING):
            assert self.manager.collection_requests == []

    def test_request_product(self):
        self.manager.request_product('name')
        with self._caplog.at_level(logging.WARNING):
            assert self.manager.product_requests == []

    def test_request_package_exclusion(self):
        self.manager.request_package_exclusion('name')
        with self._caplog.at_level(logging.WARNING):
            assert self.manager.exclude_requests == []

    def test_setup_repository_modules(self):
        self.manager.setup_repository_modules({})

    @patch('kiwi.command.Command.run')
    @patch.object(PackageManagerApt, 'process_install_requests')
    @patch('os.path.isfile')
    def test_process_install_requests_bootstrap_prebuild_root(
        self, mock_os_path_isfile, mock_process_install_requests,
        mock_Command_run
    ):
        mock_os_path_isfile.return_value = True
        self.manager.process_install_requests_bootstrap(
            bootstrap_package='bootstrap-package'
        )
        assert mock_Command_run.call_args_list == [
            call(['apt-get', '-c', 'apt.conf', '-y', 'update'], ['env']),
            call(
                [
                    'apt-get', '-c', 'apt.conf', '-y',
                    'install', 'bootstrap-package'
                ], ['env']
            ),
            call(
                [
                    'tar', '-C', 'root-dir', '-xf',
                    '/var/lib/bootstrap/bootstrap-package.{0}.tar.xz'.format(
                        defaults.PLATFORM_MACHINE
                    )
                ]
            )
        ]
        mock_process_install_requests.assert_called_once_with()
        mock_os_path_isfile.return_value = False
        with raises(KiwiFileNotFound):
            self.manager.process_install_requests_bootstrap(
                bootstrap_package='bootstrap-package'
            )

    def test_process_install_requests_bootstrap_debootstrap_no_dist(self):
        self.manager.distribution = None
        with raises(KiwiDebootstrapError):
            self.manager.process_install_requests_bootstrap()

    @patch('os.path.exists')
    def test_process_install_requests_bootstrap_debootstrap_no_script(
        self, mock_exists
    ):
        mock_exists.return_value = False
        with raises(KiwiDebootstrapError):
            self.manager.process_install_requests_bootstrap()

    @patch('kiwi.command.Command.call')
    @patch('kiwi.package_manager.apt.os.path.exists')
    @patch('kiwi.package_manager.apt.Path.wipe')
    def test_process_install_requests_bootstrap_debootstrap_failed(
        self, mock_wipe, mock_exists, mock_call
    ):
        self.manager.request_package('apt')
        mock_call.side_effect = Exception
        mock_exists.return_value = True
        mock_root_bind = Mock()
        with raises(KiwiDebootstrapError):
            self.manager.process_install_requests_bootstrap(mock_root_bind)

    @patch('kiwi.package_manager.apt.os.path.exists')
    def test_get_error_details(self, mock_exists):
        mock_exists.return_value = True
        with patch('builtins.open', create=True) as mock_open:
            file_handle = mock_open.return_value.__enter__.return_value
            file_handle.read.return_value = 'log-data'
            assert self.manager.get_error_details() == \
                file_handle.read.return_value
        mock_open.assert_called_once_with(
            'root-dir/debootstrap/debootstrap.log'
        )

    @patch('kiwi.package_manager.apt.os.path.exists')
    def test_get_error_details_no_log_file(self, mock_exists):
        mock_exists.return_value = False
        assert self.manager.get_error_details() == \
            "logfile 'root-dir/debootstrap/debootstrap.log' does not exist"

    @patch('kiwi.package_manager.apt.os.path.exists')
    def test_get_error_details_logfile_is_empty(self, mock_exists):
        mock_exists.return_value = True
        with patch('builtins.open', create=True) as mock_open:
            file_handle = mock_open.return_value.__enter__.return_value
            file_handle.read.return_value = ''
            assert self.manager.get_error_details() == \
                'logfile is empty'

    @patch('kiwi.command.Command.call')
    @patch('kiwi.package_manager.apt.Path.wipe')
    @patch('kiwi.package_manager.apt.os.path.exists')
    def test_process_install_requests_bootstrap_debootstrap(
        self, mock_exists, mock_wipe, mock_call
    ):
        self.manager.request_package('apt')
        self.manager.request_package('vim')
        call_result = Mock()
        call_result.process.communicate.return_value = ('stdout', 'stderr')
        mock_call.return_value = call_result
        mock_root_bind = Mock()
        mock_exists.return_value = True
        self.manager.process_install_requests_bootstrap(mock_root_bind)
        mock_call.assert_called_once_with(
            [
                'debootstrap', '--keyring=trusted.gpg',
                '--variant=minbase', '--include=vim',
                '--components=main,restricted', 'xenial',
                'root-dir', 'xenial_path'
            ], ['env']
        )
        assert mock_wipe.call_args_list == [
            call('root-dir/dev/fd'),
            call('root-dir/dev/pts')
        ]
        mock_root_bind.umount_kernel_file_systems.assert_called_once_with()

    def test_post_process_install_requests_bootstrap(self):
        mock_root_bind = Mock()
        self.manager.post_process_install_requests_bootstrap(mock_root_bind)
        mock_root_bind.mount_kernel_file_systems.assert_called_once_with(False)

    @patch('kiwi.command.Command.call')
    @patch('kiwi.package_manager.apt.Path.wipe')
    @patch('kiwi.package_manager.apt.os.path.exists')
    def test_process_install_requests_bootstrap_debootstrap_no_gpg_check(
        self, mock_exists, mock_wipe, mock_call
    ):
        self.manager.request_package('apt')
        self.manager.request_package('vim')
        call_result = Mock()
        call_result.process.communicate.return_value = ('stdout', 'stderr')
        mock_root_bind = Mock()
        mock_call.return_value = call_result
        mock_exists.side_effect = lambda x: True if 'xenial' in x else False
        self.manager.process_install_requests_bootstrap(mock_root_bind)
        mock_call.assert_called_once_with(
            [
                'debootstrap', '--no-check-gpg',
                '--variant=minbase', '--include=vim',
                '--components=main,restricted', 'xenial',
                'root-dir', 'xenial_path'
            ], ['env']
        )
        assert mock_wipe.call_args_list == [
            call('root-dir/dev/fd'),
            call('root-dir/dev/pts')
        ]
        mock_root_bind.umount_kernel_file_systems.assert_called_once_with()

    @patch('kiwi.command.Command.call')
    @patch('kiwi.command.Command.run')
    def test_process_install_requests(self, mock_run, mock_call):
        self.manager.request_package('vim')
        self.manager.process_install_requests()
        mock_call.assert_called_once_with([
            'chroot', 'root-dir', 'apt-get',
            '-c', 'apt.conf', '-y', 'install', 'vim'],
            ['env']
        )

    @patch('kiwi.command.Command.call')
    @patch('kiwi.command.Command.run')
    def test_process_delete_requests(self, mock_run, mock_call):
        self.manager.request_package('vim')
        self.manager.process_delete_requests()
        mock_call.assert_called_once_with(
            [
                'chroot', 'root-dir', 'apt-get', '-c', 'apt.conf', '-y',
                '--auto-remove', 'remove', 'vim'
            ], ['env']
        )

    @patch('kiwi.command.Command.call')
    @patch('kiwi.command.Command.run')
    @patch('kiwi.package_manager.apt.Path.wipe')
    @patch('glob.iglob')
    def test_process_delete_requests_force(
        self, mock_iglob, mock_Path_wipe, mock_run, mock_call
    ):
        mock_iglob.return_value = ['glob-result']
        self.manager.request_package('vim')
        self.manager.process_delete_requests(True)
        assert mock_run.call_args_list == [
            call(
                [
                    'chroot', 'root-dir', 'dpkg', '-l', 'vim'
                ]
            ),
            call(
                [
                    'cp', 'root-dir/usr/sbin/ldconfig',
                    'root-dir/usr/sbin/ldconfig.orig'
                ]
            ),
            call(
                [
                    'cp', 'root-dir/usr/bin/true',
                    'root-dir/usr/sbin/ldconfig'
                ]
            )
        ]
        mock_call.assert_called_once_with(
            [
                'chroot', 'root-dir', 'dpkg',
                '--remove', '--force-remove-reinstreq',
                '--force-remove-essential', '--force-depends', 'vim'
            ],
            ['env']
        )
        mock_iglob.call_args_list == [
            call('root-dir/var/lib/dpkg/info/vim*.pre*'),
            call('root-dir/var/lib/dpkg/info/vim*.post*')
        ]
        mock_Path_wipe.call_args_list == [
            call('glob-result'), call('glob-result')
        ]

    @patch('kiwi.command.Command.run')
    def test_post_process_delete_requests(self, mock_run):
        self.manager.post_process_delete_requests()
        assert mock_run.call_args_list == [
            call(
                [
                    'mv', 'root-dir/usr/sbin/ldconfig.orig',
                    'root-dir/usr/sbin/ldconfig'
                ]
            )
        ]

    @patch('kiwi.command.Command.run')
    def test_process_delete_requests_package_missing(self, mock_run):
        mock_run.side_effect = Exception
        self.manager.request_package('vim')
        with raises(KiwiRequestError):
            self.manager.process_delete_requests()

    @patch('kiwi.command.Command.call')
    def test_update(self, mock_call):
        self.manager.update()
        mock_call.assert_called_once_with([
            'chroot', 'root-dir', 'apt-get',
            '-c', 'apt.conf', '-y', 'upgrade'],
            ['env']
        )

    def test_process_only_required(self):
        self.manager.process_only_required()
        assert self.manager.custom_args == ['--no-install-recommends']

    def test_process_plus_recommended(self):
        self.manager.process_only_required()
        assert self.manager.custom_args == ['--no-install-recommends']
        self.manager.process_plus_recommended()
        assert '--no-install-recommends' not in self.manager.custom_args

    def test_match_package_installed(self):
        assert self.manager.match_package_installed('foo', 'Unpacking foo')

    def test_match_package_deleted(self):
        assert self.manager.match_package_deleted('foo', 'Removing foo')



from pytest import raises
from unittest.mock import patch
from kiwi.api_helper import (
    obsolete,
    decommissioned
)


@decommissioned
def method_decommissioned():
    # Implementation has been deleted
    pass


@obsolete(decommission_at='2025-01-28', version="1.2.3")
def method_obsolete():
    return 'I still exist'


class TestApiHelpers:
    def test_method_decommissioned(self):
        with raises(DeprecationWarning):
            method_decommissioned()

    @patch('warnings.warn')
    def test_method_obsolete(self, mock_warnings_warn):
        warning_message = (
            "Function 'method_obsolete' is marked obsolete "
            "since version '1.2.3' and will be decommissioned "
            "at: 2025-01-28"
        )
        assert method_obsolete() == 'I still exist'
        mock_warnings_warn.assert_called_once_with(warning_message)



from unittest.mock import (
    patch, Mock
)
from pytest import (
    raises, fixture
)

from kiwi.defaults import Defaults
from kiwi.xml_state import XMLState
from kiwi.xml_description import XMLDescription
from kiwi.bootloader.config.bootloader_spec_base import BootLoaderSpecBase
from kiwi.exceptions import KiwiKernelLookupError


class TestBootLoaderSpecBase:
    @fixture(autouse=True)
    def inject_fixtures(self, caplog):
        self._caplog = caplog

    @patch('kiwi.bootloader.config.bootloader_spec_base.FirmWare')
    def setup(self, mock_FirmWare):
        Defaults.set_platform_name('x86_64')
        description = XMLDescription(
            '../data/example_config.xml'
        )
        self.state = XMLState(
            description.load()
        )
        self.bootloader = BootLoaderSpecBase(
            self.state, 'root_dir'
        )

    @patch('kiwi.bootloader.config.bootloader_spec_base.FirmWare')
    def setup_method(self, cls, mock_FirmWare):
        self.setup()

    @patch.object(BootLoaderSpecBase, 'setup_loader')
    def test_setup_disk_image_config(self, mock_setup_loader):
        self.bootloader.get_boot_cmdline = Mock(return_value='')
        self.bootloader.setup_disk_image_config(
            'boot_uuid', 'root_uuid', 'hypervisor',
            'kernel', 'initrd', boot_options={
                'root_device': 'rootdev', 'boot_device': 'bootdev'
            }
        )
        mock_setup_loader.assert_called_once_with('disk')

    def test_setup_install_image_config(self):
        # just pass
        self.bootloader.setup_install_image_config(
            'mbrid', 'hypervisor', 'kernel', 'initrd'
        )

    def test_setup_live_image_config(self):
        # just pass
        self.bootloader.setup_live_image_config(
            'mbrid', 'hypervisor', 'kernel', 'initrd'
        )

    @patch.object(BootLoaderSpecBase, 'create_loader_image')
    def test_setup_disk_boot_images(self, mock_create_loader_image):
        self.bootloader.setup_disk_boot_images('uuid')
        mock_create_loader_image.assert_called_once_with('disk')

    @patch.object(BootLoaderSpecBase, 'create_loader_image')
    def test_setup_install_boot_images(self, mock_create_loader_image):
        self.bootloader.setup_install_boot_images('mbrid')
        mock_create_loader_image.assert_called_once_with('install(iso)')

    @patch.object(BootLoaderSpecBase, 'create_loader_image')
    def test_setup_live_boot_images(self, mock_create_loader_image):
        self.bootloader.setup_live_boot_images('mbrid')
        mock_create_loader_image.assert_called_once_with('live(iso)')

    def test_setup_loader(self):
        with raises(NotImplementedError):
            self.bootloader.setup_loader('target')

    def test_set_loader_entry(self):
        with raises(NotImplementedError):
            self.bootloader.set_loader_entry('root_dir', 'target')

    def test_create_loader_image(self):
        with raises(NotImplementedError):
            self.bootloader.create_loader_image('target')

    def test_write(self):
        # just pass
        self.bootloader.write()

    def test_setup_sysconfig_bootloader(self):
        # just pass
        self.bootloader.setup_sysconfig_bootloader()

    def test_write_meta_data(self):
        # just pass
        self.bootloader.write_meta_data()

    @patch('kiwi.bootloader.config.bootloader_spec_base.OsRelease')
    @patch('kiwi.bootloader.config.bootloader_spec_base.glob.iglob')
    def test_get_entry_name_kernel_lookup_raises(
        self, mock_iglob, mock_OsRelease
    ):
        mock_iglob.return_value = None
        with raises(KiwiKernelLookupError):
            self.bootloader.get_entry_name()

    @patch('kiwi.bootloader.config.bootloader_spec_base.OsRelease')
    @patch('kiwi.bootloader.config.bootloader_spec_base.glob.iglob')
    def test_get_entry_name_kiwi_policy(
        self, mock_iglob, mock_OsRelease
    ):
        glob_return_value = [
            ['/lib/modules/5.3.18-59.10-default'],
            []
        ]

        def get_glob(args):
            return glob_return_value.pop(0)

        mock_iglob.side_effect = get_glob
        os_release = Mock()
        os_release.get.return_value = 'opensuse-leap'
        mock_OsRelease.return_value = os_release
        assert self.bootloader.get_entry_name() == \
            'opensuse-leap-5.3.18-59.10-default.conf'

    @patch('kiwi.bootloader.config.bootloader_spec_base.OsRelease')
    @patch('kiwi.bootloader.config.bootloader_spec_base.glob.iglob')
    def test_get_entry_name_os_policy(
        self, mock_iglob, mock_OsRelease
    ):
        glob_return_value = [
            ['/lib/modules/5.3.18-59.10-default'],
            ['/boot/loader/entries/bc8499a-5.3.18-59.10-default.conf']
        ]

        def get_glob(args):
            return glob_return_value.pop(0)

        mock_iglob.side_effect = get_glob
        assert self.bootloader.get_entry_name() == \
            'bc8499a-5.3.18-59.10-default.conf'



import logging
from collections import namedtuple
from unittest.mock import (
    patch, Mock, MagicMock, mock_open
)
from pytest import (
    raises, fixture
)
import kiwi

from kiwi.builder.kis import KisBuilder
from kiwi.exceptions import KiwiKisBootImageError


class TestKisBuilder:
    @fixture(autouse=True)
    def inject_fixtures(self, caplog):
        self._caplog = caplog

    @patch('kiwi.builder.kis.FileSystemBuilder')
    @patch('kiwi.builder.kis.BootImage')
    def setup(self, mock_boot, mock_filesystem):
        self.setup = Mock()
        self.runtime_config = Mock()
        self.runtime_config.get_max_size_constraint = Mock(
            return_value=None
        )
        kiwi.builder.kis.RuntimeConfig = Mock(
            return_value=self.runtime_config
        )
        kiwi.builder.kis.SystemSetup = Mock(
            return_value=self.setup
        )
        self.boot_image_task = MagicMock()
        self.boot_image_task.boot_root_directory = 'initrd_dir'
        self.boot_image_task.initrd_filename = 'initrd_file_name'
        mock_boot.new.return_value = self.boot_image_task
        self.filesystem = MagicMock()
        self.filesystem.filename = 'myimage.fs'
        self.filesystem.root_uuid = 'some_uuid'
        mock_filesystem.return_value = self.filesystem
        self.xml_state = Mock()
        self.xml_state.profiles = None
        self.xml_state.get_image_version = Mock(
            return_value='1.2.3'
        )
        self.xml_state.get_initrd_system = Mock(
            return_value='dracut'
        )
        self.xml_state.xml_data.get_name = Mock(
            return_value='some-image'
        )
        self.xml_state.build_type = Mock()
        self.xml_state.build_type.get_kernelcmdline = Mock(
            return_value='console=hvc0'
        )
        kernel_type = namedtuple(
            'kernel', ['filename', 'version']
        )
        xen_type = namedtuple(
            'xen', ['filename', 'name']
        )
        self.kernel = Mock()
        self.kernel.get_kernel = Mock(
            return_value=kernel_type(filename='some-kernel', version='42')
        )
        self.kernel.get_xen_hypervisor = Mock(
            return_value=xen_type(filename='hypervisor', name='xen.gz')
        )
        kiwi.builder.kis.Kernel = Mock(
            return_value=self.kernel
        )
        self.kis = KisBuilder(
            self.xml_state, 'target_dir', 'root_dir',
            custom_args={'signing_keys': ['key_file_a', 'key_file_b']}
        )
        self.kis.image_name = 'myimage'
        self.kis.compressed = True

    @patch('kiwi.builder.kis.FileSystemBuilder')
    @patch('kiwi.builder.kis.BootImage')
    def setup_method(self, cls, mock_boot, mock_filesystem):
        self.setup()

    @patch('kiwi.builder.kis.FileSystemBuilder')
    @patch('kiwi.builder.kis.BootImage')
    def test_setup_warn_no_initrd_support(self, mock_boot, mock_filesystem):
        boot_image_task = MagicMock()
        boot_image_task.has_initrd_support = Mock(
            return_value=False
        )
        mock_boot.new.return_value = boot_image_task
        with self._caplog.at_level(logging.WARNING):
            KisBuilder(self.xml_state, 'target_dir', 'root_dir')

    @patch('kiwi.builder.kis.Checksum')
    @patch('kiwi.builder.kis.Compress')
    @patch('kiwi.builder.kis.ArchiveTar')
    @patch('os.rename')
    def test_create(
        self, mock_rename, mock_tar, mock_compress, mock_checksum
    ):
        tar = Mock()
        mock_tar.return_value = tar
        compress = Mock()
        mock_compress.return_value = compress
        compress.xz.return_value = 'compressed-file-name'
        checksum = Mock()
        mock_checksum.return_value = checksum
        self.boot_image_task.required = Mock(
            return_value=True
        )

        m_open = mock_open()
        with patch('builtins.open', m_open, create=True):
            self.kis.create()

        m_open.assert_called_once_with(
            'target_dir/some-image.x86_64-1.2.3.append', 'w'
        )
        m_open.return_value.write.assert_called_once_with(
            'root=UUID=some_uuid console=hvc0'
        )

        self.filesystem.create.assert_called_once_with()
        mock_rename.assert_called_once_with(
            'myimage.fs', 'myimage'
        )
        compress.xz.assert_called_once_with(None)
        checksum.md5.assert_called_once_with(
            'target_dir/some-image.x86_64-1.2.3.md5'
        )
        self.boot_image_task.prepare.assert_called_once_with()
        self.setup.export_modprobe_setup.assert_called_once_with(
            'initrd_dir'
        )
        self.boot_image_task.create_initrd.assert_called_once_with()
        self.setup.export_package_list.assert_called_once_with(
            'target_dir'
        )
        self.setup.export_package_verification.assert_called_once_with(
            'target_dir'
        )

        tar.create.assert_called_once_with('target_dir')

        self.kis.compressed = False

        m_open = mock_open()
        with patch('builtins.open', m_open, create=True):
            self.kis.create()

        m_open.return_value.write.assert_called_once_with(
            'root=UUID=some_uuid console=hvc0'
        )

        tar.create_xz_compressed.assert_called_once_with(
            'target_dir', xz_options=['--threads=0']
        )

    @patch('kiwi.builder.kis.Checksum')
    @patch('kiwi.builder.kis.Compress')
    @patch('os.rename')
    def test_create_no_kernel_found(
        self, mock_rename, mock_compress, mock_checksum
    ):
        compress = Mock()
        mock_compress.return_value = compress
        compress.compressed_filename = 'compressed-file-name'
        self.kernel.get_kernel.return_value = False
        with raises(KiwiKisBootImageError):
            self.kis.create()

    @patch('kiwi.builder.kis.Checksum')
    @patch('kiwi.builder.kis.Compress')
    @patch('os.rename')
    def test_create_no_hypervisor_found(
        self, mock_rename, mock_compress, mock_checksum
    ):
        compress = Mock()
        mock_compress.return_value = compress
        compress.compressed_filename = 'compressed-file-name'
        self.kernel.get_xen_hypervisor.return_value = False
        with raises(KiwiKisBootImageError):
            self.kis.create()



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
from string import Template
import re
import os
import logging
import glob
import shutil
from collections import OrderedDict

# project
from kiwi.bootloader.config.base import BootLoaderConfigBase
from kiwi.bootloader.template.grub2 import BootLoaderTemplateGrub2
from kiwi.command import Command
from kiwi.defaults import Defaults
from kiwi.exceptions import KiwiFileNotFound
from kiwi.firmware import FirmWare
from kiwi.path import Path
from kiwi.system.identifier import SystemIdentifier
from kiwi.utils.sync import DataSync
from kiwi.utils.sysconfig import SysConfig

from kiwi.exceptions import (
    KiwiTemplateError,
    KiwiBootLoaderGrubPlatformError,
    KiwiBootLoaderGrubModulesError,
    KiwiBootLoaderGrubSecureBootError,
    KiwiBootLoaderGrubFontError,
    KiwiDiskGeometryError
)

import kiwi.defaults as defaults

log = logging.getLogger('kiwi')


class BootLoaderConfigGrub2(BootLoaderConfigBase):
    """
    **grub2 bootloader configuration.**
    """
    def post_init(self, custom_args):
        """
        grub2 post initialization method

        :param dict custom_args:
            Contains grub config arguments

            .. code:: python

                {'grub_directory_name': 'grub|grub2'}
        """
        self.custom_args = custom_args
        self.config_options = []
        arch = Defaults.get_platform_name()
        if arch == 'x86_64':
            # grub2 support for bios and efi systems
            self.arch = arch
        elif arch.startswith('ppc64'):
            # grub2 support for ofw and opal systems
            self.arch = arch
        elif arch == 'ix86':
            # grub2 support for bios systems
            self.arch = arch
        elif arch == 'aarch64' or arch.startswith('arm'):
            # grub2 support for efi systems
            self.arch = arch
        elif arch.startswith('s390'):
            # grub2 support for s390x systems
            self.arch = arch
        elif arch == 'riscv64':
            # grub2 support for efi systems
            self.arch = arch
        else:
            raise KiwiBootLoaderGrubPlatformError(
                'host architecture %s not supported for grub2 setup' % arch
            )

        if self.custom_args and 'grub_directory_name' in self.custom_args:
            self.boot_directory_name = self.custom_args['grub_directory_name']
        else:
            self.boot_directory_name = 'grub'

        if self.custom_args and 'grub_load_command' in self.custom_args:
            self.grub_load = self.custom_args['grub_load_command']
        else:
            self.grub_load = 'source'

        if self.custom_args and 'config_options' in self.custom_args:
            self.config_options = self.custom_args['config_options']

        terminal_output = self.xml_state.get_build_type_bootloader_console()[0]
        terminal_input = self.xml_state.get_build_type_bootloader_console()[1]
        terminal_input_grub = [
            'console',
            'serial',
            'at_keyboard',
            'usb_keyboard'
        ]
        terminal_output_grub = [
            'console',
            'serial',
            'gfxterm',
            'vga_text',
            'mda_text',
            'morse',
            'spkmodem'
        ]

        self.terminal_output = \
            terminal_output if terminal_output in terminal_output_grub else 'gfxterm'
        self.terminal_input = \
            terminal_input if terminal_input in terminal_input_grub else 'console'

        self.gfxmode = self.get_gfxmode('grub2')
        self.theme = self.get_boot_theme()
        self.timeout = self.get_boot_timeout_seconds()
        self.bootpartition = self.xml_state.build_type.get_bootpartition()
        self.timeout_style = \
            self.xml_state.get_build_type_bootloader_timeout_style()
        self.displayname = self.xml_state.xml_data.get_displayname()
        self.bls = self.xml_state.get_build_type_bootloader_bls()
        self.serial_line_setup = \
            self.xml_state.get_build_type_bootloader_serial_line_setup()
        self.continue_on_timeout = self.get_continue_on_timeout()
        self.failsafe_boot = self.failsafe_boot_entry_requested()
        self.mediacheck_boot = self.xml_state.build_type.get_mediacheck()
        self.xen_guest = self.xml_state.is_xen_guest()
        self.persistency_type = \
            self.xml_state.build_type.get_devicepersistency()
        self.firmware = FirmWare(
            self.xml_state
        )
        self.target_table_type = self.firmware.get_partition_table_type()

        self.live_type = self.xml_state.build_type.get_flags()
        if not self.live_type:
            self.live_type = Defaults.get_default_live_iso_type()

        self.volume_id = self.xml_state.build_type.get_volid() or \
            Defaults.get_volume_id()
        self.install_volid = self.xml_state.build_type.get_volid() or \
            Defaults.get_install_volume_id()
        self.use_disk_password = \
            self.xml_state.get_build_type_bootloader_use_disk_password()

        self.live_boot_options = [
            'root=live:CDLABEL={0}'.format(self.volume_id),
            'rd.live.image'
        ]
        self.install_boot_options = [
            'loglevel=0'
        ]
        if self.xml_state.get_initrd_system() == 'dracut':
            self.install_boot_options.append(
                'root=install:CDLABEL={0}'.format(self.install_volid)
            )
        if self.xml_state.build_type.get_hybridpersistent():
            self.live_boot_options += \
                Defaults.get_live_iso_persistent_boot_options(
                    self.xml_state.build_type.get_hybridpersistent_filesystem()
                )

        if self.xml_state.is_xen_server():
            self.multiboot = True
        elif self.xen_guest:
            self.multiboot = False
        else:
            self.multiboot = False

        self.grub2 = BootLoaderTemplateGrub2()
        self.config = None
        self.efi_boot_path = None
        self.cmdline_failsafe = None
        self.root_reference = None
        self.cmdline = None
        self.iso_boot = False
        self.early_boot_script_efi = None

    def write(self):
        """
        Write bootloader configuration

        * writes grub.cfg template by KIWI if template system is used
        * creates an embedded fat efi image for EFI ISO boot
        """
        if self.config:
            log.info('Writing KIWI template grub.cfg file')
            config_dir = self._get_grub2_boot_path()
            config_file = config_dir + '/grub.cfg'
            Path.create(config_dir)
            with open(config_file, 'w') as config:
                config.write(self.config)

    def setup_sysconfig_bootloader(self):
        raise NotImplementedError

    def write_meta_data(
        self, root_device=None, write_device=None, boot_options=''
    ):
        """
        Write bootloader setup meta data files

        * cmdline arguments initialization
        * etc/default/grub setup file
        * etc/default/zipl2grub.conf.in (s390 only)
        * etc/sysconfig/bootloader

        :param string root_device: root device node
        :param string write_device: overlay root write device node
        :param string boot_options: kernel options as string
        :param bool iso_boot: indicate target is an ISO
        """
        self.cmdline = ' '.join(
            [self.get_boot_cmdline(root_device, write_device), boot_options]
        )
        self.cmdline_failsafe = ' '.join(
            [self.cmdline, Defaults.get_failsafe_kernel_options(), boot_options]
        )
        self.root_reference = self._get_root_cmdline_parameter(
            root_device
        )

        self._setup_default_grub()
        self._setup_sysconfig_bootloader()
        if self.arch.startswith('s390'):
            self._setup_zipl2grub_conf()

    def setup_disk_image_config(
        self, boot_uuid=None, root_uuid=None, hypervisor=None,
        kernel=None, initrd=None, boot_options={}
    ):
        """
        Create grub2 config file to boot from disk using grub2-mkconfig

        :param string boot_uuid: unused
        :param string root_uuid: unused
        :param string hypervisor: unused
        :param string kernel: unused
        :param string initrd: unused
        :param dict boot_options:

        options dictionary that has to contain the root and boot
        device and optional volume configuration. KIWI has to
        mount the system prior to run grub2-mkconfig.

        .. code:: python

            {
                'root_device': string,
                'boot_device': string,
                'efi_device': string,
                'system_volumes':
                    volume_manager_instance.get_volumes(),
                'system_root_volume':
                    volume_manager_instance.get_root_volume_name()
            }
        """
        self._mount_system(
            boot_options.get('root_device'),
            boot_options.get('boot_device'),
            boot_options.get('efi_device'),
            boot_options.get('system_volumes'),
            boot_options.get('system_root_volume')
        )
        config_file = os.sep.join(
            [
                self.root_mount.mountpoint, 'boot',
                self.boot_directory_name, 'grub.cfg'
            ]
        )
        Command.run(
            [
                'chroot', self.root_mount.mountpoint,
                os.path.basename(self._get_grub2_mkconfig_tool()), '-o',
                config_file.replace(self.root_mount.mountpoint, '')
            ]
        )
        if boot_options.get('root_device') != boot_options.get('boot_device'):
            # Create a boot -> . link on the boot partition.
            # The link is useful if the grub mkconfig command
            # references all boot files from /boot/... even if
            # an extra boot partition should cause it to read the
            # data from the toplevel
            bash_command = [
                'cd', os.sep.join([self.root_mount.mountpoint, 'boot']), '&&',
                'rm', '-f', 'boot', '&&',
                'ln', '-s', '.', 'boot'
            ]
            # not every filesystem supports symlinks, and the link is optional
            # therefore don't fail in case the link cannot be created
            Command.run(
                ['bash', '-c', ' '.join(bash_command)], raise_on_error=False
            )

        # Patch the written grub config file to actually work:
        # Unfortunately the grub tooling has several bugs and issues
        # which prevents it to work in image build environments.
        # More details can be found in the individual methods.
        # One fine day the following fix methods can hopefully
        # be deleted...
        if self.xml_state.build_type.get_overlayroot_write_partition() is not False:
            self._fix_grub_root_device_reference(config_file, boot_options)
            self._fix_grub_loader_entries_boot_cmdline()
            self._fix_grub_loader_entries_linux_and_initrd_paths()

        if self.firmware.efi_mode() and self.early_boot_script_efi:
            self._copy_grub_config_to_efi_path(
                self.efi_mount.mountpoint, self.early_boot_script_efi
            )

    def setup_install_image_config(
        self, mbrid, hypervisor='xen.gz', kernel='linux', initrd='initrd'
    ):
        """
        Create grub2 config file to boot from an ISO install image

        :param string mbrid: mbrid file name on boot device
        :param string hypervisor: hypervisor name
        :param string kernel: kernel name
        :param string initrd: initrd name
        """
        log.info('Creating grub2 install config file from template')
        self.iso_boot = True
        has_graphics = False
        has_serial = False
        if 'gfxterm' in self.terminal_output:
            has_graphics = True
        if 'serial' in self.terminal_output or 'serial' in self.terminal_input:
            has_serial = True
        parameters = {
            'search_params': '--file --set=root /boot/' + mbrid.get_id(),
            'default_boot': self.get_install_image_boot_default(),
            'kernel_file': kernel,
            'initrd_file': initrd,
            'boot_options': ' '.join(
                [self.cmdline] + self.install_boot_options
            ),
            'failsafe_boot_options': ' '.join(
                [self.cmdline_failsafe] + self.install_boot_options
            ),
            'gfxmode': self.gfxmode,
            'theme': self.theme,
            'boot_timeout': self.timeout,
            'boot_timeout_style': self.timeout_style or 'menu',
            'serial_line_setup': self.serial_line_setup or '',
            'title': self.get_menu_entry_install_title(),
            'bootpath': self.get_boot_path('iso'),
            'boot_directory_name': self.boot_directory_name,
            'efi_image_name': Defaults.get_efi_image_name(self.arch),
            'terminal_input': self.terminal_input,
            'terminal_output': self.terminal_output
        }
        custom_template_path = self._get_custom_template()
        if custom_template_path:
            log.info('--> Using custom boot template')
            with open(custom_template_path) as custom_template_file:
                template = Template(custom_template_file.read())
        elif self.multiboot:
            log.info('--> Using multiboot install template')
            parameters['hypervisor'] = hypervisor
            template = self.grub2.get_multiboot_install_template(
                self.failsafe_boot, has_graphics, has_serial,
                self.continue_on_timeout
            )
        else:
            log.info('--> Using standard boot install template')
            template = self.grub2.get_install_template(
                self.failsafe_boot, has_graphics, has_serial,
                self.continue_on_timeout
            )
        try:
            self.config = template.substitute(parameters)
        except Exception as e:
            raise KiwiTemplateError(
                '%s: %s' % (type(e).__name__, format(e))
            )
        if self.firmware.efi_mode() and self.early_boot_script_efi:
            self._copy_grub_config_to_efi_path(
                self.boot_dir, self.early_boot_script_efi, 'iso'
            )

    def setup_live_image_config(
        self, mbrid: SystemIdentifier, hypervisor='xen.gz', kernel='linux', initrd='initrd'
    ):
        """
        Create grub2 config file to boot a live media ISO image

        :param string mbrid: mbrid file name on boot device
        :param string hypervisor: hypervisor name
        :param string kernel: kernel name
        :param string initrd: initrd name
        """
        log.info('Creating grub2 live ISO config file from template')
        self.iso_boot = True
        has_graphics = False
        has_serial = False
        if 'gfxterm' in self.terminal_output:
            has_graphics = True
        if 'serial' in self.terminal_output or 'serial' in self.terminal_input:
            has_serial = True
        parameters = {
            'search_params': '--file --set=root /boot/' + mbrid.get_id(),
            'default_boot': '0',
            'kernel_file': kernel,
            'initrd_file': initrd,
            'boot_options': ' '.join(
                [self.cmdline] + self.live_boot_options
            ),
            'failsafe_boot_options': ' '.join(
                [self.cmdline_failsafe] + self.live_boot_options
            ),
            'gfxmode': self.gfxmode,
            'theme': self.theme,
            'boot_timeout': self.timeout,
            'boot_timeout_style': self.timeout_style or 'menu',
            'serial_line_setup': self.serial_line_setup or '',
            'title': self.get_menu_entry_title(plain=True),
            'bootpath': self.get_boot_path('iso'),
            'boot_directory_name': self.boot_directory_name,
            'efi_image_name': Defaults.get_efi_image_name(self.arch),
            'terminal_input': self.terminal_input,
            'terminal_output': self.terminal_output
        }
        custom_template_path = self._get_custom_template()
        if custom_template_path:
            log.info('--> Using custom boot template')
            with open(custom_template_path) as custom_template_file:
                template = Template(custom_template_file.read())
        elif self.multiboot:
            log.info('--> Using multiboot template')
            parameters['hypervisor'] = hypervisor
            template = self.grub2.get_multiboot_iso_template(
                self.failsafe_boot, has_graphics, has_serial,
                self.mediacheck_boot
            )
        else:
            log.info('--> Using standard boot template')
            template = self.grub2.get_iso_template(
                self.failsafe_boot, has_graphics, has_serial,
                self.mediacheck_boot
            )
        try:
            self.config = template.substitute(parameters)
        except Exception as e:
            raise KiwiTemplateError(
                '%s: %s' % (type(e).__name__, format(e))
            )
        if self.firmware.efi_mode() and self.early_boot_script_efi:
            self._copy_grub_config_to_efi_path(
                self.boot_dir, self.early_boot_script_efi, 'iso'
            )

    def setup_install_boot_images(self, mbrid: SystemIdentifier, lookup_path: str = None) -> None:
        """
        Create/Provide grub2 boot images and metadata

        In order to boot from the ISO grub2 modules, images and theme
        data needs to be created and provided at the correct place on
        the iso filesystem

        :param string mbrid: mbrid file name on boot device
        :param string lookup_path: custom module lookup path
        """
        log.info('Creating grub2 bootloader images')
        self.efi_boot_path = self.create_efi_path(in_sub_dir='')

        log.info('--> Creating identifier file %s', mbrid.get_id())
        grub_boot_path = self._get_grub2_boot_path()
        Path.create(
            grub_boot_path
        )
        mbrid.write(
            self.boot_dir + '/boot/' + mbrid.get_id()
        )
        mbrid.write(
            self.boot_dir + '/boot/mbrid'
        )

        self._copy_theme_data_to_boot_directory(lookup_path, 'iso')

        if self._supports_bios_modules():
            self._copy_bios_modules_to_boot_directory(lookup_path)
            self._setup_bios_image(mbrid=mbrid, lookup_path=lookup_path)

        if self.firmware.efi_mode():
            self._setup_EFI_path(lookup_path)

        if self.firmware.efi_mode() == 'efi':
            self._setup_efi_image(mbrid=mbrid, lookup_path=lookup_path, target_type='iso')
            self._copy_efi_modules_to_boot_directory(lookup_path)
        elif self.firmware.efi_mode() == 'uefi':
            self._setup_secure_boot_efi_image(
                lookup_path=lookup_path, mbrid=mbrid, target_type='iso'
            )

        log.info('--> Creating loopback config')
        loopback_file = os.path.normpath(
            os.sep.join([grub_boot_path, 'loopback.cfg'])
        )
        self._create_loopback_config(
            loopback_file
        )

    def setup_live_boot_images(self, mbrid: SystemIdentifier, lookup_path=None):
        """
        Create/Provide grub2 boot images and metadata

        Calls setup_install_boot_images because no different action required
        """
        self.setup_install_boot_images(mbrid, lookup_path)

    def setup_disk_boot_images(self, boot_uuid, lookup_path=None):
        """
        Create/Provide grub2 boot images and metadata

        In order to boot from the disk grub2 modules, images and theme
        data needs to be created and provided at the correct place in
        the filesystem

        :param string boot_uuid: boot device UUID
        :param string lookup_path: custom module lookup path
        """
        log.info('Creating grub2 bootloader images')

        if self.firmware.efi_mode():
            self.efi_boot_path = self.create_efi_path()

        self._copy_theme_data_to_boot_directory(lookup_path, 'disk')

        if not self.xen_guest and self._supports_bios_modules():
            self._copy_bios_modules_to_boot_directory(lookup_path)

        if self.firmware.efi_mode() == 'efi':
            self._setup_efi_image(uuid=boot_uuid, lookup_path=lookup_path)
            self._copy_efi_modules_to_boot_directory(lookup_path)
        elif self.firmware.efi_mode() == 'uefi':
            self._copy_efi_modules_to_boot_directory(lookup_path)
            if not self._get_shim_install():
                self._setup_secure_boot_efi_image(
                    lookup_path=lookup_path, uuid=boot_uuid
                )

        if self.xen_guest:
            self._copy_xen_modules_to_boot_directory(lookup_path)

    def _copy_grub_config_to_efi_path(
        self, root_path, config_file, target_type='disk'
    ):
        efi_boot_path = Defaults.get_shim_vendor_directory(
            root_path
        )
        if not efi_boot_path:
            # not all distributors installs a vendor directory but
            # have them in their encoded early boot script. Thus
            # the following code tries to look up the vendor string
            # from the signed grub binary
            grub_image = Defaults.get_signed_grub_loader(self.root_dir, target_type)
            if grub_image and grub_image.filename:
                bash_command = [
                    'strings', grub_image.filename, '|', 'grep', r'EFI\/'
                ]
                efi_boot_path = Command.run(
                    ['bash', '-c', ' '.join(bash_command)],
                    raise_on_error=False
                ).output
                if efi_boot_path:
                    efi_boot_path = os.path.normpath(
                        os.sep.join([root_path, efi_boot_path.strip()])
                    )
        if not efi_boot_path:
            efi_boot_path = os.path.normpath(
                os.sep.join([root_path, 'EFI/BOOT'])
            )
        Path.create(efi_boot_path)
        grub_config_file_for_efi_boot = os.sep.join(
            [efi_boot_path, 'grub.cfg']
        )
        if config_file != grub_config_file_for_efi_boot:
            log.info(
                'Copying {0} -> {1} to be found by EFI'.format(
                    config_file, grub_config_file_for_efi_boot
                )
            )
            shutil.copy(
                config_file, grub_config_file_for_efi_boot
            )

    def _supports_bios_modules(self):
        if self.arch == 'ix86' or self.arch == 'x86_64':
            return True
        return False

    def _setup_sysconfig_bootloader(self):
        """
        Create or update etc/sysconfig/bootloader by the following
        parameters required according to the grub2 bootloader setup

        * LOADER_TYPE
        * LOADER_LOCATION
        * DEFAULT_APPEND
        * FAILSAFE_APPEND
        * SECURE_BOOT
        * TRUSTED_BOOT
        """
        sysconfig_bootloader_entries = {
            'LOADER_TYPE':
                'grub2-efi' if self.firmware.efi_mode() else 'grub2',
            'LOADER_LOCATION':
                'none' if self.firmware.efi_mode() else 'mbr'
        }
        if '--set-trusted-boot' in self.config_options:
            sysconfig_bootloader_entries['TRUSTED_BOOT'] = 'yes'
        if self.firmware.efi_mode() == 'uefi':
            sysconfig_bootloader_entries['SECURE_BOOT'] = 'yes'
        elif self.firmware.efi_mode() == 'efi':
            sysconfig_bootloader_entries['SECURE_BOOT'] = 'no'
        if self.cmdline:
            sysconfig_bootloader_entries['DEFAULT_APPEND'] = '"{0}"'.format(
                self.cmdline
            )
        if self.cmdline_failsafe:
            sysconfig_bootloader_entries['FAILSAFE_APPEND'] = '"{0}"'.format(
                self.cmdline_failsafe
            )

        sysconfig_bootloader_location = ''.join(
            [self.root_dir, '/etc/sysconfig/']
        )
        if os.path.exists(sysconfig_bootloader_location):
            log.info('Writing sysconfig bootloader file')
            sysconfig_bootloader_file = ''.join(
                [sysconfig_bootloader_location, 'bootloader']
            )
            sysconfig_bootloader = SysConfig(
                sysconfig_bootloader_file
            )
            sysconfig_bootloader_entries_sorted = OrderedDict(
                sorted(sysconfig_bootloader_entries.items())
            )
            for key, value in list(sysconfig_bootloader_entries_sorted.items()):
                log.info('--> {0}:{1}'.format(key, value))
                sysconfig_bootloader[key] = value
            sysconfig_bootloader.write()

    def _setup_zipl2grub_conf(self):
        zipl2grub_config_file = ''.join(
            [self.root_dir, '/etc/default/zipl2grub.conf.in']
        )
        zipl2grub_config_file_orig = ''.join(
            [self.root_dir, '/etc/default/zipl2grub.conf.in.orig']
        )
        target_type = self.xml_state.get_build_type_bootloader_targettype()
        target_blocksize = self.xml_state.build_type.get_target_blocksize()
        if target_type and os.path.exists(zipl2grub_config_file):
            if os.path.exists(zipl2grub_config_file_orig):
                # reset the original template file first
                shutil.copy(zipl2grub_config_file_orig, zipl2grub_config_file)
            else:
                # no copy of the original template, create it
                shutil.copy(zipl2grub_config_file, zipl2grub_config_file_orig)
            with open(zipl2grub_config_file) as zipl_config_file:
                zipl_config = zipl_config_file.read()
                zipl_config = re.sub(
                    r'(:menu)',
                    ':menu\n'
                    '    targettype = {0}\n'
                    '    targetbase = {1}\n'
                    '    targetblocksize = {2}\n'
                    '    targetoffset = {3}\n'
                    '    {4}'.format(
                        target_type,
                        self.custom_args['targetbase'],
                        target_blocksize or 512,
                        self._get_partition_start(
                            self.custom_args['targetbase']
                        ),
                        self._get_disk_geometry(
                            self.custom_args['targetbase']
                        )
                    ),
                    zipl_config
                )
            log.debug('Updated zipl template as follows')
            with open(zipl2grub_config_file, 'w') as zipl_config_file:
                log.debug(zipl_config)
                zipl_config_file.write(zipl_config)

    def _setup_default_grub(self):
        """
        Create or update etc/default/grub by parameters required
        according to the root filesystem setup

        * GRUB_DEFAULT
        * GRUB_TIMEOUT
        * GRUB_TIMEOUT_STYLE
        * SUSE_BTRFS_SNAPSHOT_BOOTING
        * SUSE_REMOVE_LINUX_ROOT_PARAM
        * GRUB_BACKGROUND
        * GRUB_THEME
        * GRUB_SERIAL_COMMAND
        * GRUB_CMDLINE_LINUX
        * GRUB_CMDLINE_LINUX_DEFAULT
        * GRUB_GFXMODE
        * GRUB_TERMINAL_INPUT
        * GRUB_TERMINAL_OUTPUT
        * GRUB_DISTRIBUTOR
        * GRUB_DISABLE_LINUX_UUID
        * GRUB_DISABLE_LINUX_PARTUUID
        * GRUB_ENABLE_LINUX_LABEL
        """
        # elements to set only if not already in etc/default/grub
        grub_default_if_not_set_entries = {
            'GRUB_DEFAULT': 'saved'
        }
        # elements to set in any case
        grub_default_entries = {
            'GRUB_TIMEOUT': self.timeout,
            'GRUB_GFXMODE': self.gfxmode
        }
        if self.terminal_input:
            grub_default_entries['GRUB_TERMINAL_INPUT'] = '"{0}"'.format(
                self.terminal_input
            )
        if self.terminal_output:
            grub_default_entries['GRUB_TERMINAL_OUTPUT'] = '"{0}"'.format(
                self.terminal_output
            )
        grub_final_cmdline = re.sub(
            r'(^root=[^\s]+)|( root=[^\s]+)', '', self.cmdline
        ).strip()
        if self.persistency_type == 'by-label':
            label = re.search(r'(^root=[^\s]+)|( root=[^\s]+)', self.cmdline)
            if label:
                for match in label.groups():
                    if match and 'LABEL' in match:
                        grub_default_entries['GRUB_CMDLINE_LINUX'] = \
                            '"{0}"'.format(match.strip())
                        grub_default_entries['SUSE_REMOVE_LINUX_ROOT_PARAM'] = \
                            'true'
            grub_default_entries['GRUB_ENABLE_LINUX_LABEL'] = 'true'
            grub_default_entries['GRUB_DISABLE_LINUX_UUID'] = 'true'
        elif self.persistency_type == 'by-partuuid':
            grub_default_entries['GRUB_DISABLE_LINUX_UUID'] = 'true'
            grub_default_entries['GRUB_DISABLE_LINUX_PARTUUID'] = 'false'
        if self.displayname:
            grub_default_entries['GRUB_DISTRIBUTOR'] = '"{0}"'.format(
                self.displayname
            )
        if self.timeout_style:
            grub_default_entries['GRUB_TIMEOUT_STYLE'] = self.timeout_style
        if grub_final_cmdline:
            grub_default_entries['GRUB_CMDLINE_LINUX_DEFAULT'] = '"{0}"'.format(
                grub_final_cmdline
            )
        if self.serial_line_setup and \
           ('serial' in self.terminal_input or 'serial' in self.terminal_output):
            grub_default_entries['GRUB_SERIAL_COMMAND'] = '"{0}"'.format(
                self.serial_line_setup
            )
        if self.theme:
            theme_setup = '{0}/{1}/theme.txt'
            grub_default_entries['GRUB_THEME'] = theme_setup.format(
                ''.join(['/boot/', self.boot_directory_name, '/themes']),
                self.theme
            )
            theme_background = '{0}/{1}/background.png'.format(
                ''.join(['/boot/', self.boot_directory_name, '/themes']),
                self.theme
            )
            if os.path.exists(os.sep.join([self.root_dir, theme_background])):
                grub_default_entries['GRUB_BACKGROUND'] = theme_background
        if self.xml_state.build_type.get_btrfs_root_is_snapshot():
            grub_default_entries['SUSE_BTRFS_SNAPSHOT_BOOTING'] = 'true'
        if self.custom_args.get('crypto_disk'):
            grub_default_entries['GRUB_ENABLE_CRYPTODISK'] = 'y'

        enable_blscfg_implemented = Command.run(
            [
                'grep', '-q', 'GRUB_ENABLE_BLSCFG',
                self._get_grub2_mkconfig_tool()
            ], raise_on_error=False
        )
        if self.bls and enable_blscfg_implemented.returncode == 0:
            grub_default_entries['GRUB_ENABLE_BLSCFG'] = 'true'

        if grub_default_entries:
            log.info('Writing grub2 defaults file')
            grub_default_location = ''.join([self.root_dir, '/etc/default/'])
            if os.path.exists(grub_default_location):
                grub_default_file = ''.join([grub_default_location, 'grub'])
                grub_default = SysConfig(grub_default_file)
                grub_default_entries_sorted = OrderedDict(
                    sorted(grub_default_entries.items())
                )
                for key, value in list(grub_default_entries_sorted.items()):
                    log.info('--> {0}:{1}'.format(key, value))
                    grub_default[key] = value
                grub_default_if_not_set_entries_sorted = OrderedDict(
                    sorted(grub_default_if_not_set_entries.items())
                )
                for key, value in list(grub_default_if_not_set_entries_sorted.items()):
                    if not grub_default.get(key):
                        log.info('--> {0}:{1}'.format(key, value))
                        grub_default[key] = value
                grub_default.write()

    def _setup_secure_boot_efi_image(
        self, lookup_path, uuid=None, mbrid=None, target_type='disk'
    ):
        """
        Provide the shim loader and the shim signed grub2 loader
        in the required boot path. Normally this task is done by
        the shim-install tool. However, shim-install does not exist
        on all distributions and the script does not operate well
        in e.g CD environments from which we generate live and/or
        install media. Thus shim-install is used if possible at
        install time of the bootloader because it requires access
        to the target block device. In any other case this setup
        code should act as the fallback solution
        """
        log.warning(
            'Running fallback setup for shim secure boot efi image'
        )
        if not lookup_path:
            lookup_path = self.boot_dir
        grub_image = Defaults.get_signed_grub_loader(lookup_path, target_type)
        if not grub_image:
            raise KiwiBootLoaderGrubSecureBootError(
                'Signed grub2 efi loader not found'
            )
        shim_image = Defaults.get_shim_loader(lookup_path)
        if shim_image and shim_image.filename:
            # The shim concept is based on a two step system including a
            # grub image(shim) that got signed by Microsoft followed by
            # a grub image that got signed by the shim. The shim image
            # is the one that gets loaded by the firmware which itself
            # loads the second stage grub image
            target_efi_image_name = self._get_efi_image_name()
            target_grub_image_name = os.sep.join(
                [self.efi_boot_path, grub_image.binaryname]
            )
            if not os.path.isfile(target_efi_image_name):
                log.info(
                    f'--> Using shim image: {shim_image.filename}'
                )
                Command.run(
                    ['cp', shim_image.filename, target_efi_image_name]
                )
            if not os.path.isfile(target_grub_image_name):
                log.info(
                    f'--> Using grub image: {grub_image.filename}'
                )
                Command.run(
                    ['cp', grub_image.filename, target_grub_image_name]
                )
            mok_manager = Defaults.get_mok_manager(lookup_path)
            if mok_manager:
                target_mok_manager = os.sep.join(
                    [self.efi_boot_path, os.path.basename(mok_manager)]
                )
                if not os.path.isfile(target_mok_manager):
                    log.info(
                        f'--> Using mok image: {mok_manager}'
                    )
                    Command.run(
                        ['cp', mok_manager, self.efi_boot_path]
                    )
        else:
            # Without shim a self signed grub image is used that
            # gets loaded by the firmware
            target_efi_image_name = self._get_efi_image_name()
            if not os.path.isfile(target_efi_image_name):
                log.info(
                    f'--> No shim image, using grub image: {grub_image.filename}'
                )
                Command.run(
                    ['cp', grub_image.filename, target_efi_image_name]
                )
        self._create_efi_config_search(uuid, mbrid)

    def _setup_efi_image(
        self, uuid=None, mbrid=None, lookup_path=None, target_type='disk'
    ):
        """
        Provide the unsigned grub2 efi image in the required boot path
        If a prebuilt efi image as provided by the distribution could
        be found, this image will be used. If no such image could be
        found we create a custom image with a pre defined set of
        grub modules
        """
        if not lookup_path:
            lookup_path = self.boot_dir
        grub_image = Defaults.get_unsigned_grub_loader(lookup_path, target_type)
        if grub_image and self.xml_state.build_type.get_overlayroot_write_partition() is not False:
            log.info(f'--> Using prebuilt unsigned efi image: {grub_image}')
            Command.run(
                ['cp', grub_image, self._get_efi_image_name()]
            )
            self._create_efi_config_search(uuid, mbrid)
        else:
            log.info('--> Creating unsigned efi image')
            self._create_efi_image(
                uuid, mbrid, lookup_path
            )

    def _setup_bios_image(self, mbrid, lookup_path=None):
        """
        Provide bios grub image
        """
        if not lookup_path:
            lookup_path = self.boot_dir
        grub_image = Defaults.get_grub_bios_core_loader(lookup_path)
        if grub_image:
            log.info(f'--> Using prebuilt bios image: {grub_image}')
        else:
            log.info('--> Creating bios image')
            self._create_bios_image(
                mbrid, lookup_path
            )
        bash_command = ' '.join(
            [
                'cat',
                self._get_bios_modules_path(lookup_path) + '/cdboot.img',
                grub_image or self._get_bios_image_name(lookup_path),
                '>',
                os.sep.join(
                    [
                        self._get_bios_modules_path(lookup_path),
                        Defaults.get_iso_grub_loader()
                    ]
                )
            ]
        )
        Command.run(
            ['bash', '-c', bash_command]
        )

    def _create_embedded_fat_efi_image(self, path):
        """
        Creates a EFI system partition image at the given path.
        Must be called after setup_install_boot_images and write.
        """
        fat_image_mbsize = int(
            self.xml_state.build_type
                .get_efifatimagesize() or defaults.EFI_FAT_IMAGE_SIZE
        )
        Command.run(
            ['qemu-img', 'create', path, f'{fat_image_mbsize}M']
        )
        Command.run(
            ['mkdosfs', '-n', 'BOOT', path]
        )
        Command.run(
            [
                'mcopy', '-Do', '-s', '-i', path,
                self.boot_dir + '/EFI', '::'
            ]
        )

    def _create_efi_image(self, uuid=None, mbrid=None, lookup_path=None):
        early_boot_script = os.path.normpath(
            os.sep.join([self.efi_boot_path, 'earlyboot.cfg'])
        )
        self.early_boot_script_efi = early_boot_script
        if uuid:
            self._create_early_boot_script_for_uuid_search(
                early_boot_script, uuid
            )
        else:
            self._create_early_boot_script_for_mbrid_search(
                early_boot_script, mbrid
            )
        root_efi_path = '/boot/efi/EFI/BOOT/'
        root_efi_image = root_efi_path + Defaults.get_efi_image_name(self.arch)

        Path.create(self.root_dir + root_efi_path)

        module_list = Defaults.get_grub_efi_modules(multiboot=self.xen_guest)
        module_path = self._get_efi_modules_path(self.root_dir)
        if os.path.exists(module_path + '/linuxefi.mod'):
            module_list.append('linuxefi')

        early_boot_script_on_media = \
            self.root_dir + root_efi_path + 'earlyboot.cfg'
        if early_boot_script != early_boot_script_on_media:
            log.debug(
                f'Copy earlyboot to media path: {early_boot_script_on_media}'
            )
            shutil.copy(
                early_boot_script, early_boot_script_on_media
            )
        mkimage_call = Command.run(
            [
                'chroot', self.root_dir,
                os.path.basename(
                    self._get_grub2_mkimage_tool()
                ) or 'grub2-mkimage',
                '-O', Defaults.get_efi_module_directory_name(self.arch),
                '-o', root_efi_image,
                '-c', root_efi_path + 'earlyboot.cfg',
                '-p', os.path.join(self.get_boot_path(), self.boot_directory_name),
                '-d', module_path.replace(self.root_dir, '')
            ] + module_list
        )
        log.debug(mkimage_call.output)

        # Copy generated EFI image to the media directory if this
        # is different from the system root directory, e.g the case
        # for live image builds or when using a custom kiwi initrd
        efi_image_root_file = self.root_dir + root_efi_image
        efi_image_media_file = os.sep.join(
            [self.efi_boot_path, os.path.basename(efi_image_root_file)]
        )
        if (efi_image_root_file != efi_image_media_file):
            log.debug(
                f'Copy grub image to media path: {efi_image_media_file}'
            )
            Path.create(os.path.dirname(efi_image_media_file))
            shutil.copy(efi_image_root_file, efi_image_media_file)

    def _create_efi_config_search(self, uuid=None, mbrid=None):
        efi_boot_config = self.efi_boot_path + '/grub.cfg'
        self.early_boot_script_efi = efi_boot_config
        if uuid:
            self._create_early_boot_script_for_uuid_search(
                efi_boot_config, uuid
            )
        else:
            self._create_early_boot_script_for_mbrid_search(
                efi_boot_config, mbrid
            )

    def _create_bios_image(self, mbrid, lookup_path=None):
        early_boot_script = os.path.normpath(
            os.sep.join([self._get_grub2_boot_path(), 'earlyboot.cfg'])
        )
        self._create_early_boot_script_for_mbrid_search(
            early_boot_script, mbrid
        )
        early_boot_script_on_media = os.sep.join(
            [self.root_dir, 'boot', self.boot_directory_name, 'earlyboot.cfg']
        )
        if early_boot_script != early_boot_script_on_media:
            log.debug(
                f'Copy earlyboot to media path: {early_boot_script_on_media}'
            )
            Path.create(
                os.path.dirname(early_boot_script_on_media)
            )
            shutil.copy(
                early_boot_script, early_boot_script_on_media
            )
        mkimage_call = Command.run(
            [
                'chroot', self.root_dir,
                os.path.basename(
                    self._get_grub2_mkimage_tool()
                ) or 'grub2-mkimage',
                '-O', Defaults.get_bios_module_directory_name(),
                '-o', self._get_bios_image_name(self.root_dir).replace(
                    self.root_dir, ''
                ),
                '-c', early_boot_script.replace(self.boot_dir, ''),
                '-p', os.sep.join(
                    [self.get_boot_path(), self.boot_directory_name]
                ),
                '-d', self._get_bios_modules_path(self.root_dir).replace(
                    self.root_dir, ''
                )
            ] + Defaults.get_grub_bios_modules(multiboot=self.xen_guest)
        )
        log.debug(mkimage_call.output)

        # Copy generated EFI image to the media directory if this
        # is different from the system root directory, e.g the case
        # for live image builds or when using a custom kiwi initrd
        bios_image_root_file = self._get_bios_image_name(self.root_dir)
        bios_image_media_file = bios_image_root_file.replace(
            self.root_dir, lookup_path
        )
        if (bios_image_root_file != bios_image_media_file):
            log.debug(
                f'Copy grub image to media path: {bios_image_media_file}'
            )
            Path.create(os.path.dirname(bios_image_media_file))
            shutil.copy(bios_image_root_file, bios_image_media_file)

    def _create_early_boot_script_for_uuid_search(self, filename, uuid):
        with open(filename, 'w') as early_boot:
            early_boot.write(
                'set btrfs_relative_path="yes"{0}'.format(os.linesep)
            )
            if self.custom_args.get('boot_is_crypto'):
                early_boot.write(
                    'insmod cryptodisk{0}'.format(os.linesep)
                )
                early_boot.write(
                    'insmod luks{0}'.format(os.linesep)
                )
                early_boot.write(
                    'cryptomount -u {0}{1}'.format(
                        uuid.replace('-', ''), os.linesep
                    )
                )
                early_boot.write(
                    'set root="cryptouuid/{0}"{1}'.format(
                        uuid.replace('-', ''), os.linesep
                    )
                )
            early_boot.write(
                'search --fs-uuid --set=root {0}{1}'.format(uuid, os.linesep)
            )
            early_boot.write(
                'set prefix=($root){0}{1}'.format(
                    os.path.join(self.get_boot_path(), self.boot_directory_name), os.linesep
                )
            )
            early_boot.write(
                '{0} ($root){1}/grub.cfg{2}'.format(self.grub_load, os.path.join(
                    self.get_boot_path(), self.boot_directory_name
                ), os.linesep)
            )

    def _create_early_boot_script_for_mbrid_search(self, filename, mbrid):
        with open(filename, 'w') as early_boot:
            early_boot.write(
                'set btrfs_relative_path="yes"{0}'.format(os.linesep)
            )
            if mbrid:
                early_boot.write(
                    f'search --file --set=root /boot/{mbrid.get_id()}{os.linesep}'
                )
            else:
                # Fallback search for /boot/mbrid
                early_boot.write(
                    f'search --file --set=root /boot/mbrid{os.linesep}'
                )
            early_boot.write(
                'set prefix=($root)/boot/{0}{1}'.format(
                    self.boot_directory_name, os.linesep
                )
            )
            early_boot.write(
                '{0} ($root)/boot/{1}/grub.cfg{2}'.format(
                    self.grub_load, self.boot_directory_name, os.linesep
                )
            )

    def _get_grub2_mkimage_tool(self):
        for grub_mkimage_tool in ['grub2-mkimage', 'grub-mkimage']:
            grub_mkimage_file_path = Path.which(
                grub_mkimage_tool, root_dir=self.root_dir
            )
            if grub_mkimage_file_path:
                return grub_mkimage_file_path

    def _get_grub2_mkconfig_tool(self):
        for grub_mkconfig_tool in ['grub2-mkconfig', 'grub-mkconfig']:
            grub_mkconfig_file_path = Path.which(
                grub_mkconfig_tool, root_dir=self.root_dir
            )
            if grub_mkconfig_file_path:
                return grub_mkconfig_file_path

    def _get_grub2_boot_path(self):
        return self.boot_dir + '/boot/' + self.boot_directory_name

    def _get_efi_image_name(self):
        return os.sep.join(
            [
                self.efi_boot_path,
                Defaults.get_efi_image_name(self.arch)
            ]
        )

    def _get_bios_image_name(self, lookup_path):
        return os.sep.join(
            [
                self._get_bios_modules_path(lookup_path),
                Defaults.get_bios_image_name()
            ]
        )

    def _get_efi_modules_path(self, lookup_path=None):
        return self._get_module_path(
            Defaults.get_efi_module_directory_name(self.arch),
            lookup_path
        )

    def _get_bios_modules_path(self, lookup_path=None):
        return self._get_module_path('i386-pc', lookup_path)

    def _get_xen_modules_path(self, lookup_path=None):
        return self._get_module_path(
            Defaults.get_efi_module_directory_name('x86_64_xen'),
            lookup_path
        )

    def _get_module_path(self, format_name, lookup_path=None):
        if not lookup_path:
            lookup_path = self.boot_dir
        return Defaults.get_grub_path(lookup_path, format_name)

    def _find_theme_background_file(self, lookup_path):
        background_pattern = os.sep.join(
            [
                lookup_path, 'boot', self.boot_directory_name, 'themes',
                '*', 'background.png'
            ]
        )
        for background_file in glob.iglob(background_pattern):
            return background_file

    def _copy_theme_data_to_boot_directory(self, lookup_path, target):
        if not lookup_path:
            lookup_path = self.boot_dir
        font_name = 'unicode.pf2'
        efi_font_dir = Defaults.get_grub_efi_font_directory(
            lookup_path
        )
        boot_fonts_dir = os.path.normpath(
            os.sep.join(
                [
                    self.boot_dir,
                    self.get_boot_path(target),
                    self.boot_directory_name,
                    'fonts'
                ]
            )
        )
        try:
            unicode_font = Defaults.get_grub_path(
                lookup_path, font_name
            )
            if not os.path.exists(os.sep.join([boot_fonts_dir, font_name])):
                Path.create(boot_fonts_dir)
                Command.run(
                    ['cp', unicode_font, boot_fonts_dir]
                )
            if efi_font_dir:
                Command.run(
                    ['cp', unicode_font, efi_font_dir]
                )
        except Exception as issue:
            raise KiwiBootLoaderGrubFontError(
                'Setting up unicode font failed with {0}'.format(issue)
            )

        boot_theme_dir = os.sep.join(
            [self.boot_dir, 'boot', self.boot_directory_name, 'themes']
        )
        Path.create(boot_theme_dir)

        if self.theme:
            theme_dir = Defaults.get_grub_path(
                lookup_path, 'themes/' + self.theme, raise_on_error=False
            )
            boot_theme_background_file = self._find_theme_background_file(
                lookup_path
            )
            if theme_dir and os.path.exists(theme_dir):
                if boot_theme_background_file:
                    # A background file was found. Preserve a copy of the
                    # file which was created at install time of the theme
                    # package by the activate-theme script
                    boot_theme_background_backup_file = os.sep.join(
                        [self.boot_dir, 'background.png']
                    )
                    Command.run(
                        [
                            'cp', boot_theme_background_file,
                            boot_theme_background_backup_file
                        ]
                    )
                # sync theme data from install path to boot path
                data = DataSync(
                    theme_dir, boot_theme_dir
                )
                data.sync_data(
                    options=['-a']
                )
                if boot_theme_background_file:
                    # Install preserved background file to the theme
                    Command.run(
                        [
                            'mv', boot_theme_background_backup_file,
                            os.sep.join([boot_theme_dir, self.theme])
                        ]
                    )
            elif boot_theme_background_file:
                # assume all theme data is in the directory of the
                # background file and just sync that directory to the
                # boot path
                data = DataSync(
                    os.path.dirname(boot_theme_background_file), boot_theme_dir
                )
                data.sync_data(
                    options=['-a']
                )

        self._check_boot_theme_exists()

    def _check_boot_theme_exists(self):
        if self.theme:
            theme_dir = os.sep.join(
                [
                    self.boot_dir, 'boot', self.boot_directory_name,
                    'themes', self.theme
                ]
            )
            if not os.path.exists(theme_dir):
                log.warning('Theme %s not found', theme_dir)
                log.warning('Set bootloader terminal to console mode')
                self.terminal_input = 'console'
                self.terminal_output = 'console'

    def _setup_EFI_path(self, lookup_path):
        """
        Copy efi boot data from lookup_path to the root directory
        """
        if not lookup_path:
            lookup_path = self.boot_dir
        efi_path = lookup_path + '/boot/efi/'
        if os.path.exists(efi_path):
            efi_data = DataSync(efi_path, self.boot_dir)
            efi_data.sync_data(options=['-a'])

    def _copy_efi_modules_to_boot_directory(self, lookup_path):
        self._copy_modules_to_boot_directory_from(
            self._get_efi_modules_path(lookup_path)
        )

    def _copy_bios_modules_to_boot_directory(self, lookup_path):
        self._copy_modules_to_boot_directory_from(
            self._get_bios_modules_path(lookup_path)
        )

    def _copy_xen_modules_to_boot_directory(self, lookup_path):
        self._copy_modules_to_boot_directory_from(
            self._get_xen_modules_path(lookup_path)
        )

    def _copy_modules_to_boot_directory_from(self, module_path):
        boot_module_path = \
            self._get_grub2_boot_path() + '/' + os.path.basename(module_path)
        try:
            data = DataSync(
                module_path + '/', boot_module_path
            )
            data.sync_data(
                options=['-a'], exclude=['*.module']
            )
        except Exception as e:
            raise KiwiBootLoaderGrubModulesError(
                'Module synchronisation failed with: %s' % format(e)
            )

    def _get_shim_install(self):
        return Path.which(
            filename='shim-install', root_dir=self.boot_dir
        )

    def _fix_grub_root_device_reference(self, config_file, boot_options):
        if self.root_reference:
            # grub2-mkconfig has no idea how the correct root= setup is
            # for disk images created with overlayroot enabled or in a
            # buildservice worker environment. Because of that the mkconfig
            # tool just finds the raw partition loop device and includes it
            # which is wrong. In this particular case we have to patch the
            # written config file and replace the wrong root= reference with
            # the correct value.
            with open(config_file) as grub_config_file:
                grub_config = grub_config_file.read()
                # The following expression matches any of the following
                # grub mkconfig root= settings and replaces it with a
                # correct value
                # 1. root=LOCAL-KIWI-MAPPED-DEVICE
                # 2. root=[a-zA-Z]=ANY-LINUX-BY-ID-VALUE
                grub_config = re.sub(
                    r'(root=[a-zA-Z]+=[a-zA-Z0-9:\.-]+)|(root={0})'.format(
                        boot_options.get('root_device')
                    ),
                    '{0}'.format(self.root_reference),
                    grub_config
                )

            with open(config_file, 'w') as grub_config_file:
                grub_config_file.write(grub_config)

            if self.firmware.efi_mode():
                vendor_grubenv_file = \
                    Defaults.get_vendor_grubenv(self.efi_mount.mountpoint)
                if vendor_grubenv_file:
                    with open(vendor_grubenv_file) as vendor_grubenv:
                        grubenv = vendor_grubenv.read()
                        grubenv = grubenv.replace(
                            'root={0}'.format(
                                boot_options.get('root_device')
                            ),
                            self.root_reference
                        )
                    with open(vendor_grubenv_file, 'w') as vendor_grubenv:
                        vendor_grubenv.write(grubenv)

    def _fix_grub_loader_entries_boot_cmdline(self):
        if self.cmdline:
            # For distributions that follows the bootloader spec here:
            # https://www.freedesktop.org/wiki/Specifications/BootLoaderSpec
            # the menu entries are managed in extra files written to
            # boot/loader/entries. The grub2-mkconfig tool imports the
            # information from there and has no own menu entries data
            # in grub.cfg anymore. Unfortunately the system that writes
            # those new boot/loader/entries has several issues. It produces
            # a complete mess when used in chroot environments. In such
            # an environment it takes the proc/cmdline data from the host
            # that built the image and completely ignores the setup of
            # the image description. Overall for image building the
            # tooling is completely broken and we are forced to patch the
            # entire cmdline for the menuentries on such systems.
            loader_entries_pattern = os.sep.join(
                [
                    self.root_mount.mountpoint,
                    'boot', 'loader', 'entries', '*.conf'
                ]
            )
            for menu_entry_file in glob.iglob(loader_entries_pattern):
                with open(menu_entry_file) as grub_menu_entry_file:
                    menu_entry = grub_menu_entry_file.read()
                    menu_entry = re.sub(
                        r'options (.*)',
                        r'options {0}'.format(self.cmdline),
                        menu_entry
                    )
                with open(menu_entry_file, 'w') as grub_menu_entry_file:
                    grub_menu_entry_file.write(menu_entry)

    def _fix_grub_loader_entries_linux_and_initrd_paths(self):
        # For the same reasons encoded in _fix_grub_loader_entries_boot_cmdline
        # this method exists. In this method the wrong paths to the linux
        # kernel and initrd gets fixed
        loader_entries_pattern = os.sep.join(
            [
                self.root_mount.mountpoint,
                'boot', 'loader', 'entries', '*.conf'
            ]
        )
        for menu_entry_file in glob.iglob(loader_entries_pattern):
            with open(menu_entry_file) as grub_menu_entry_file:
                menu_entry = grub_menu_entry_file.read().split(os.linesep)

            for line_number, menu_entry_line in enumerate(menu_entry):
                if bool(re.match(r'^(linux|initrd) .*', menu_entry_line)):

                    log.debug(f'Existing loader entry: {menu_entry_line}')
                    config_path = menu_entry_line.split(' ', 1)[0]

                    basename = os.path.basename(menu_entry_line)
                    if self.bootpartition:
                        config_path = (f'{config_path} /{basename}')
                    else:
                        config_path = (f'{config_path} /boot/{basename}')

                    menu_entry[line_number] = config_path
                    log.debug(f'Updated loader entry: {config_path}')

            menu_entry = os.linesep.join(menu_entry)
            with open(menu_entry_file, 'w') as grub_menu_entry_file:
                grub_menu_entry_file.write(menu_entry)

    def _get_partition_start(self, disk_device):
        if self.target_table_type == 'dasd':
            blocks = self._get_dasd_disk_geometry_element(
                disk_device, 'blocks per track'
            )
            bash_command = [
                'fdasd', '-f', '-s', '-p', disk_device,
                '|', 'grep', '"^ "',
                '|', 'head', '-n', '1',
                '|', 'tr', '-s', '" "'
            ]
            fdasd_call = Command.run(
                ['bash', '-c', ' '.join(bash_command)]
            )
            fdasd_output = fdasd_call.output
            try:
                start_track = int(fdasd_output.split(' ')[2].lstrip())
            except Exception:
                raise KiwiDiskGeometryError(
                    'unknown partition format: %s' % fdasd_output
                )
            return '{0}'.format(start_track * blocks)
        else:
            bash_command = ' '.join(
                [
                    'sfdisk', '--dump', disk_device,
                    '|', 'grep', '"1 :"',
                    '|', 'cut', '-f1', '-d,',
                    '|', 'cut', '-f2', '-d='
                ]
            )
            return Command.run(
                ['bash', '-c', bash_command]
            ).output.strip()

    def _get_disk_geometry(self, disk_device):
        disk_geometry = ''
        if self.target_table_type == 'dasd':
            disk_geometry = 'targetgeometry = {0},{1},{2}'.format(
                self._get_dasd_disk_geometry_element(
                    disk_device, 'cylinders'
                ),
                self._get_dasd_disk_geometry_element(
                    disk_device, 'tracks per cylinder'
                ),
                self._get_dasd_disk_geometry_element(
                    disk_device, 'blocks per track'
                )
            )
        return disk_geometry

    def _get_dasd_disk_geometry_element(self, disk_device, search):
        fdasd = ['fdasd', '-f', '-p', disk_device]
        bash_command = fdasd + ['|', 'grep', '"' + search + '"']
        fdasd_call = Command.run(
            ['bash', '-c', ' '.join(bash_command)]
        )
        fdasd_output = fdasd_call.output
        try:
            return int(fdasd_output.split(':')[1].lstrip())
        except Exception:
            raise KiwiDiskGeometryError(
                'unknown format for disk geometry: %s' % fdasd_output
            )

    def _create_loopback_config(self, filename):
        with open(filename, 'w') as loopback_cfg:
            loopback_cfg.write(
                'source /boot/{0}/grub.cfg{1}'.format(
                    self.boot_directory_name, os.linesep
                )
            )

    def _get_custom_template(self) -> str:
        if not self.xml_state.build_type.bootloader:
            return ''
        template_file = self.xml_state.build_type.bootloader[0].get_grub_template()
        if not template_file:
            return ''

        template_path = os.path.join(os.path.abspath(self.xml_state.xml_data.description_dir),
                                     self.xml_state.build_type.bootloader[0].get_grub_template())
        if not os.path.exists(template_path):
            raise KiwiFileNotFound('failed to locate custom GRUB template %s' % template_file)

        return template_path



from unittest.mock import Mock
from pytest import raises

from kiwi.bootloader.install.zipl import BootLoaderInstallZipl


class TestBootLoaderInstallZipl:
    def setup(self):
        self.bootloader = BootLoaderInstallZipl(
            'root_dir', Mock()
        )

    def setup_method(self, cls):
        self.setup()

    def test_install(self):
        with raises(NotImplementedError):
            self.bootloader.install()

    def test_install_required(self):
        assert self.bootloader.install_required() is False

    def test_secure_boot_install(self):
        # just pass
        self.bootloader.secure_boot_install()



import sys

from unittest.mock import (
    patch, call, Mock, MagicMock
)

import kiwi

from ..test_helper import argv_kiwi_tests

from kiwi.tasks.image_info import ImageInfoTask

from collections import namedtuple


class TestImageInfoTask:
    def setup(self):
        sys.argv = [
            sys.argv[0], 'image', 'info',
            '--description', '../data/image_info',
            '--resolve-package-list'
        ]
        result_type = namedtuple(
            'result_type', [
                'uri', 'installsize_bytes', 'arch', 'version', 'checksum'
            ]
        )
        self.result_info = {
            'image': 'LimeJeOS',
            'resolved-packages': {
                'packagename': {
                    'version': '0.8.15', 'arch': 'x86_64',
                    'source': 'uri', 'status': 'added_by_dependency_solver',
                    'installsize_bytes': 42
                },
                'filesystem': {
                    'version': '42', 'arch': 'x86_64',
                    'source': 'uri', 'status': 'listed_in_kiwi_description',
                    'installsize_bytes': 0
                }
            }
        }
        self.runtime_checker = Mock()
        kiwi.tasks.base.RuntimeChecker = Mock(
            return_value=self.runtime_checker
        )
        self.runtime_config = Mock()
        kiwi.tasks.base.RuntimeConfig = Mock(
            return_value=self.runtime_config
        )
        self.solver = MagicMock()
        self.solver.solve = Mock(
            return_value={
                'packagename': result_type(
                    uri='uri', installsize_bytes=42,
                    arch='x86_64', version='0.8.15', checksum='checksum'
                ),
                'filesystem': result_type(
                    uri='uri', installsize_bytes=0,
                    arch='x86_64', version='42', checksum='checksum'
                )
            }
        )
        kiwi.tasks.image_info.Sat = Mock(
            return_value=self.solver
        )

        self.solver_repository = Mock()
        kiwi.tasks.image_info.SolverRepository = Mock(
            return_value=self.solver_repository
        )

        kiwi.tasks.image_info.Help = Mock(
            return_value=Mock()
        )
        self.task = ImageInfoTask()

    def setup_method(self, cls):
        self.setup()

    def teardown(self):
        sys.argv = argv_kiwi_tests

    def teardown_method(self, cls):
        self.teardown()

    def _init_command_args(self):
        self.task.command_args = {}
        self.task.command_args['help'] = False
        self.task.command_args['info'] = False
        self.task.command_args['--description'] = '../data/image_info'
        self.task.command_args['--add-repo'] = []
        self.task.command_args['--ignore-repos'] = False
        self.task.command_args['--resolve-package-list'] = False
        self.task.command_args['--print-xml'] = False
        self.task.command_args['--print-yaml'] = False
        self.task.command_args['--print-toml'] = False

    @patch('kiwi.tasks.image_info.DataOutput')
    def test_process_image_info(self, mock_out):
        self._init_command_args()
        self.task.command_args['info'] = True
        self.task.process()
        self.runtime_checker.check_repositories_configured.assert_called_once_with()
        mock_out.assert_called_once_with(
            {'image': 'LimeJeOS'}
        )

    @patch('kiwi.tasks.image_info.DataOutput')
    def test_process_image_info_color_output(self, mock_out):
        self._init_command_args()
        self.task.command_args['info'] = True
        self.task.global_args['--color-output'] = True
        self.task.process()
        mock_out.assert_called_once_with(
            {'image': 'LimeJeOS'}, style='color'
        )

    @patch('kiwi.tasks.image_info.DataOutput')
    @patch('kiwi.tasks.image_info.SolverRepository.new')
    @patch('kiwi.tasks.image_info.Uri')
    @patch('kiwi.tasks.image_info.SolverRepositoryBase.get_repo_type')
    def test_process_image_info_resolve_package_list(
        self, mock_get_repo_type, mock_uri, mock_solver_repo_new, mock_out
    ):
        uri = Mock()
        uri.uri = 'http://example.org/some/path'
        mock_uri.return_value = uri
        mock_get_repo_type.return_value = 'apt-deb'
        self.solver.set_dist_type.return_value = {
            'arch': 'amd64'
        }
        self._init_command_args()
        self.task.command_args['info'] = True
        self.task.command_args['--resolve-package-list'] = True
        self.task.process()

        assert self.solver.add_repository.called
        assert mock_uri.call_args_list == [
            call(uri='http://us.archive.ubuntu.com/ubuntu/', source_type=''),
            call(
                'http://us.archive.ubuntu.com/ubuntu/dists/focal/'
                'main/binary-amd64', 'apt-deb', ''
            ),
            call(
                'http://us.archive.ubuntu.com/ubuntu/dists/focal/'
                'multiverse/binary-amd64', 'apt-deb', ''
            ),
            call(
                'http://us.archive.ubuntu.com/ubuntu/dists/focal/'
                'restricted/binary-amd64', 'apt-deb', ''
            ),
            call(
                'http://us.archive.ubuntu.com/ubuntu/dists/focal/'
                'universe/binary-amd64', 'apt-deb', ''
            ),
            call('obs://Devel:PubCloud:AmazonEC2/SLE_12_GA', 'rpm-md', '')
        ]
        mock_out.assert_called_once_with(self.result_info)

    @patch('kiwi.xml_state.XMLState.add_repository')
    @patch('kiwi.tasks.image_info.DataOutput')
    def test_process_image_info_add_repo(self, mock_out, mock_state):
        self._init_command_args()
        self.task.command_args['--add-repo'] = [
            'http://example.com,rpm-md,alias'
        ]
        self.task.process()
        mock_state.assert_called_once_with(
            'http://example.com', 'rpm-md', 'alias', None
        )

    @patch('kiwi.xml_state.XMLState.delete_repository_sections')
    @patch('kiwi.tasks.image_info.DataOutput')
    def test_process_image_info_delete_repos(self, mock_out, mock_delete_repos):
        self._init_command_args()
        self.task.command_args['--ignore-repos'] = True
        self.task.process()
        mock_delete_repos.assert_called_once_with()

    def test_process_image_info_help(self):
        self._init_command_args()
        self.task.command_args['help'] = True
        self.task.command_args['info'] = True
        self.task.process()
        self.task.manual.show.assert_called_once_with(
            'kiwi::image::info'
        )

    @patch('kiwi.tasks.image_info.DataOutput')
    def test_process_image_info_print_xml(self, mock_out):
        self._init_command_args()
        self.task.command_args['--print-xml'] = True
        self.task.process()
        tmpfile, message = mock_out.display_file.call_args.args
        assert tmpfile.startswith('/var/tmp/kiwi_xslt-')
        assert message == 'Description(XML):'

    @patch('kiwi.tasks.image_info.DataOutput')
    def test_process_image_info_print_yaml(self, mock_out):
        self._init_command_args()
        self.task.command_args['--print-yaml'] = True
        self.task.process()
        tmpfile, message = mock_out.display_file.call_args.args
        assert tmpfile.startswith('/var/tmp/kiwi_xslt-')
        assert message == 'Description(YAML):'

    @patch('kiwi.tasks.image_info.DataOutput')
    def test_process_image_info_print_toml(self, mock_out):
        self._init_command_args()
        self.task.command_args['--print-toml'] = True
        self.task.process()
        tmpfile, message = mock_out.display_file.call_args.args
        assert tmpfile.startswith('/var/tmp/kiwi_xslt-')
        assert message == 'Description(TOML):'



import os
from unittest.mock import (
    patch, MagicMock, Mock, call
)
from pytest import fixture

from kiwi.system.mount import ImageSystem
from kiwi.storage.mapped_device import MappedDevice


class TestImageSystem:
    @fixture(autouse=True)
    def inject_fixtures(self, caplog):
        self._caplog = caplog

    @patch('os.path.exists')
    def setup(self, mock_os_path_exists):
        mock_os_path_exists.return_value = True
        self.device_map = {
            'root': MappedDevice('/dev/root-device', Mock()),
            'readonly': MappedDevice('/dev/readonly-root-device', Mock()),
            'boot': MappedDevice('/dev/boot-device', Mock()),
            'efi': MappedDevice('/dev/efi-device', Mock()),
        }
        self.volumes = {
            'name': {
                'volume_options': 'a,b,c',
                'volume_device': '/dev/vgroup/volume'
            }
        }
        self.image_system = ImageSystem(
            self.device_map, 'root_dir', self.volumes
        )

    @patch('os.path.exists')
    def setup_method(self, cls, mock_os_path_exists):
        self.setup()

    def test_mountpoint(self):
        some_mount = MagicMock()
        self.image_system.mount_list.append(some_mount)
        assert self.image_system.mountpoint() == some_mount.mountpoint

    @patch('kiwi.system.mount.MountManager')
    def test_mount(self, mock_MountManager):
        self.image_system.mount()
        root_mount_mountpoint = self.image_system.mount_list[0].mountpoint
        assert mock_MountManager.call_args_list == [
            call(device='/dev/readonly-root-device'),
            call(
                device='/dev/boot-device',
                mountpoint=os.path.join(root_mount_mountpoint, 'boot')
            ),
            call(
                device='/dev/efi-device',
                mountpoint=os.path.join(root_mount_mountpoint, 'boot', 'efi')
            ),
            call(
                device='/dev/vgroup/volume',
                mountpoint=os.path.join(root_mount_mountpoint, 'name')
            ),
            call(
                device='root_dir/image',
                mountpoint=os.path.join(root_mount_mountpoint, 'image')
            ),
            call(
                device='tmpfs',
                mountpoint=os.path.join(root_mount_mountpoint, 'tmp')
            ),
            call(
                device='tmpfs',
                mountpoint=os.path.join(root_mount_mountpoint, 'var', 'tmp')
            ),
            call(
                device='/proc',
                mountpoint=os.path.join(root_mount_mountpoint, 'proc')
            ),
            call(
                device='/sys',
                mountpoint=os.path.join(root_mount_mountpoint, 'sys')
            ),
            call(
                device='/dev',
                mountpoint=os.path.join(root_mount_mountpoint, 'dev')
            )
        ]

    @patch('kiwi.system.mount.MountManager')
    def test_mount_s390(self, mock_MountManager):
        self.image_system.arch = 's390x'
        self.image_system.mount()
        root_mount_mountpoint = self.image_system.mount_list[0].mountpoint
        assert mock_MountManager.call_args_list[1] == call(
            device='/dev/boot-device',
            mountpoint=os.path.join(root_mount_mountpoint, 'boot', 'zipl')
        )

    def test_umount(self):
        some_mount = MagicMock()
        some_mount.is_mounted.return_value = True
        self.image_system.mount_list.append(some_mount)
        self.image_system.umount()
        some_mount.umount.assert_called_once_with()

    @patch('os.path.exists')
    @patch('kiwi.system.mount.MountManager')
    def test_context_manager_exit(self, mock_MountManager, mock_os_path_exists):
        mount = MagicMock()
        mount.is_mounted.return_value = True
        mock_os_path_exists.return_value = True
        mock_MountManager.return_value = mount
        device_map = {
            'root': MappedDevice('/dev/root-device', Mock())
        }
        with ImageSystem(device_map, 'root_dir') as system:
            system.mount()
        mount.umount.assert_called()



# Copyright (c) 2024 SUSE Software Solutions Germany GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
from pytest import raises
from unittest.mock import Mock

from kiwi.bootloader.config.custom import BootLoaderConfigCustom


class TestBootLoaderConfigCustom:
    def setup(self):
        self.bootloader = BootLoaderConfigCustom(
            Mock(), 'root_dir'
        )

    def setup_method(self, cls):
        self.setup()

    def test_setup_disk_boot_images(self):
        with raises(NotImplementedError):
            self.bootloader.setup_disk_boot_images('0815')

    def test_setup_disk_image_config(self):
        with raises(NotImplementedError):
            self.bootloader.setup_disk_image_config(Mock())

    def test_setup_install_boot_images(self):
        with raises(NotImplementedError):
            self.bootloader.setup_install_boot_images('0815')

    def test_setup_install_image_config(self):
        with raises(NotImplementedError):
            self.bootloader.setup_install_image_config(Mock())

    def test_setup_live_boot_images(self):
        with raises(NotImplementedError):
            self.bootloader.setup_live_boot_images('0815')

    def test_setup_live_image_config(self):
        with raises(NotImplementedError):
            self.bootloader.setup_live_image_config(Mock())

    def test_setup_sysconfig_bootloader(self):
        with raises(NotImplementedError):
            self.bootloader.setup_sysconfig_bootloader()

    def test_write(self):
        with raises(NotImplementedError):
            self.bootloader.write()



import logging
from unittest.mock import patch

import sys
from unittest.mock import MagicMock
from pytest import fixture

from .test_helper import argv_kiwi_tests

from kiwi.defaults import Defaults
from kiwi.defaults import grub_loader_type
from kiwi.defaults import shim_loader_type


class TestDefaults:
    @fixture(autouse=True)
    def inject_fixtures(self, caplog):
        self._caplog = caplog

    def setup(self):
        self.defaults = Defaults()

    def setup_method(self, cls):
        self.setup()

    def teardown(self):
        sys.argv = argv_kiwi_tests

    def teardown_method(self, cls):
        self.teardown()

    def test_get(self):
        assert self.defaults.get('kiwi_align') == 1048576
        assert self.defaults.get('kiwi_startsector') == 2048
        assert self.defaults.get('kiwi_sectorsize') == 512
        assert self.defaults.get('kiwi_inode_size') == 256
        assert self.defaults.get('kiwi_inode_ratio') == 16384
        assert self.defaults.get('kiwi_min_inodes') == 20000
        assert self.defaults.get('kiwi_revision')

    def test_to_profile(self):
        profile = MagicMock()
        self.defaults.to_profile(profile)
        profile.add.assert_any_call('kiwi_align', 1048576)
        profile.add.assert_any_call('kiwi_startsector', 2048)
        profile.add.assert_any_call('kiwi_sectorsize', 512)
        profile.add.assert_any_call(
            'kiwi_revision', self.defaults.get('kiwi_revision')
        )

    def test_get_preparer(self):
        assert Defaults.get_preparer() == \
            'KIWI - https://github.com/OSInside/kiwi'

    def test_get_publisher(self):
        assert Defaults.get_publisher() == 'SUSE LINUX GmbH'

    def test_get_default_shared_cache_location(self):
        assert Defaults.get_shared_cache_location() == 'var/cache/kiwi'

    @patch('kiwi.defaults.Path.which')
    def test_get_grub_boot_directory_name(self, mock_which):
        mock_which.return_value = 'grub2-install-was-found'
        assert Defaults.get_grub_boot_directory_name(
            lookup_path='lookup_path'
        ) == 'grub2'
        mock_which.return_value = None
        assert Defaults.get_grub_boot_directory_name(
            lookup_path='lookup_path'
        ) == 'grub'

    def test_get_live_dracut_modules_from_flag(self):
        assert Defaults.get_live_dracut_modules_from_flag('foo') == \
            ['kiwi-live']
        assert Defaults.get_live_dracut_modules_from_flag('overlay') == \
            ['kiwi-live']
        assert Defaults.get_live_dracut_modules_from_flag('dmsquash') == \
            ['dmsquash-live', 'livenet']

    def test_get_iso_boot_path(self):
        Defaults.set_platform_name('i686')
        assert Defaults.get_iso_boot_path() == 'boot/ix86'
        Defaults.set_platform_name('x86_64')
        assert Defaults.get_iso_boot_path() == 'boot/x86_64'

    @patch('kiwi.defaults.glob.iglob')
    def test_get_unsigned_grub_loader(self, mock_glob):
        mock_glob.return_value = ['/usr/share/grub2/x86_64-efi/grub.efi']
        assert Defaults.get_unsigned_grub_loader('root') == \
            mock_glob.return_value.pop()
        mock_glob.assert_called_once_with('root/usr/share/grub*/*-efi/grub.efi')
        mock_glob.reset_mock()
        mock_glob.return_value = []
        assert Defaults.get_unsigned_grub_loader('root') is None

    def test_is_x86_arch(self):
        assert Defaults.is_x86_arch('x86_64') is True
        assert Defaults.is_x86_arch('aarch64') is False

    @patch('os.path.exists')
    def test_get_vendor_grubenv(self, mock_path_exists):
        mock_path_exists.return_value = True
        assert Defaults.get_vendor_grubenv('boot/efi') == \
            'boot/efi/EFI/fedora/grubenv'

    def test_get_exclude_list_from_custom_exclude_files(self):
        assert Defaults.get_exclude_list_from_custom_exclude_files(
            '../data/root-dir'
        ) == [
            'usr/bin/qemu-binfmt',
            'usr/bin/qemu-x86_64-binfmt',
            'usr/bin/qemu-x86_64'
        ]

    @patch('yaml.safe_load')
    def test_get_exclude_list_from_custom_exclude_files_is_invalid(
        self, mock_yaml_safe_load
    ):
        mock_yaml_safe_load.return_value = {'invalid': 'artificial'}
        with self._caplog.at_level(logging.WARNING):
            assert Defaults.get_exclude_list_from_custom_exclude_files(
                '../data/root-dir'
            ) == []

    def test_get_exclude_list_for_root_data_sync(self):
        assert Defaults.get_exclude_list_for_root_data_sync() == [
            'image', '.profile', '.kconfig',
            'run/*', 'tmp/*',
            '.buildenv', 'var/cache/kiwi'
        ]
        assert Defaults.get_exclude_list_for_root_data_sync(no_tmpdirs=False) == [
            'image', '.profile', '.kconfig',
            '.buildenv', 'var/cache/kiwi'
        ]

    @patch('glob.iglob')
    def test_get_signed_grub_loader(self, mock_iglob):
        def iglob_no_matches(pattern):
            return []

        def iglob_simple_match(pattern):
            if '/boot/efi/EFI/*/grub*.efi' in pattern:
                return ['root_path/boot/efi/EFI/BOOT/grub.efi']
            else:
                return []

        def iglob_custom_binary_match(pattern):
            if '/usr/lib/grub/x86_64-efi-signed/grubx64.efi.signed' in pattern:
                return [
                    'root_path'
                    '/usr/lib/grub/x86_64-efi-signed/grubx64.efi.signed'
                ]
            else:
                return []

        mock_iglob.side_effect = iglob_no_matches
        assert Defaults.get_signed_grub_loader('root_path') is None

        mock_iglob.side_effect = iglob_simple_match
        assert Defaults.get_signed_grub_loader('root_path') == grub_loader_type(
            filename='root_path/boot/efi/EFI/BOOT/grub.efi',
            binaryname='grub.efi'
        )

        mock_iglob.side_effect = iglob_custom_binary_match
        assert Defaults.get_signed_grub_loader('root_path') == grub_loader_type(
            filename='root_path'
            '/usr/lib/grub/x86_64-efi-signed/grubx64.efi.signed',
            binaryname='grubx64.efi'
        )

    @patch('glob.iglob')
    def test_get_mok_manager(self, mock_iglob):
        mock_iglob.return_value = []
        assert Defaults.get_mok_manager('root_path') is None

        mock_iglob.return_value = ['some_glob_result']
        assert Defaults.get_mok_manager('root_path') == 'some_glob_result'

    @patch('glob.iglob')
    def test_get_shim_loader(self, mock_iglob):
        def iglob_simple_match(pattern):
            if '/usr/lib64/efi/shim.efi' in pattern:
                return ['root_path/usr/lib64/efi/shim.efi']
            else:
                return []

        def iglob_custom_binary_match(pattern):
            if '/usr/lib/shim/shimx64.efi.signed' in pattern:
                return [
                    'root_path'
                    '/usr/lib/shim/shimx64.efi.signed'
                ]
            else:
                return []

        assert Defaults.get_shim_loader('root_path') is None

        mock_iglob.side_effect = iglob_simple_match
        assert Defaults.get_shim_loader('root_path') == shim_loader_type(
            filename='root_path'
            '/usr/lib64/efi/shim.efi',
            binaryname='shim.efi'
        )

        mock_iglob.side_effect = iglob_custom_binary_match
        assert Defaults.get_shim_loader('root_path/usr/lib/shim/shimx64.efi.signed') == shim_loader_type(
            filename='root_path'
            '/usr/lib/shim/shimx64.efi.signed',
            binaryname='shimx64.efi'
        )

    @patch('os.path.exists')
    def test_get_snapper_config_template_file(self, mock_os_path_exists):
        mock_os_path_exists.return_value = False
        assert Defaults.get_snapper_config_template_file('root') == ''
        mock_os_path_exists.return_value = True
        assert Defaults.get_snapper_config_template_file('root') == \
            'root/etc/snapper/config-templates/default'



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import logging


class LoggerSchedulerFilter(logging.Filter):
    """
    **Extended standard logging Filter**
    """
    def filter(self, record: logging.LogRecord) -> bool:
        """
        Messages from apscheduler scheduler instances are filtered out
        They conflict with console progress information

        :param tuple record: logging message record

        :return: True|False

        :rtype: bool
        """
        ignorables = [
            'apscheduler.scheduler',
            'apscheduler.executors.default'
        ]
        return bool(record.name not in ignorables)


class InfoFilter(logging.Filter):
    """
    **Extended standard logging Filter**
    """
    def filter(self, record: logging.LogRecord) -> bool:
        """
        Only messages with record level INFO can pass
        for messages with another level an extra handler is used

        :param tuple record: logging message record

        :return: True|False

        :rtype: bool
        """
        return True if record.levelno == logging.INFO else False


class DebugFilter(logging.Filter):
    """
    **Extended standard debug logging Filter**
    """
    def filter(self, record: logging.LogRecord) -> bool:
        """
        Only messages with record level DEBUG can pass
        for messages with another level an extra handler is used

        :param tuple record: logging message record

        :return: True|False

        :rtype: bool
        """
        return True if record.levelno == logging.DEBUG else False


class ErrorFilter(logging.Filter):
    """
    **Extended standard error logging Filter**
    """
    def filter(self, record: logging.LogRecord) -> bool:
        """
        Only messages with record level DEBUG can pass
        for messages with another level an extra handler is used

        :param tuple record: logging message record

        :return: True|False

        :rtype: bool
        """
        return True if record.levelno == logging.ERROR else False


class WarningFilter(logging.Filter):
    """
    **Extended standard warning logging Filter**
    """
    def filter(self, record: logging.LogRecord) -> bool:
        """
        Only messages with record level WARNING can pass
        for messages with another level an extra handler is used

        :param tuple record: logging message record

        :return: True|False

        :rtype: bool
        """
        return True if record.levelno == logging.WARNING else False
